'''Tests for PipelineStep and friends.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pytest

from ...algorithms.connect import ConnectorModel
from .pipeline_step import AlreadyDefinedError, PipelineStep
from .pipeline_template import PipelineTemplate

# pylint: disable=missing-function-docstring,protected-access,duplicate-code


def test_name_sunny_day() -> None:
    pipe = PipelineTemplate(name='foo')
    dut = pipe.step()
    assert dut.pipeline_designator_component == '_'
    dut.set_name('Some_Designator')
    assert dut.pipeline_designator_component == 'some_designator'


def test_name_already_defined() -> None:
    pipe = PipelineTemplate(name='foo')
    dut = pipe.step()
    dut.set_name('Some_Designator')
    with pytest.raises(AlreadyDefinedError, match=r'another_designator.*some_designator'):
        dut.set_name('Another_Designator')


def test_nominate_chains() -> None:
    pipe = PipelineTemplate(name='foo')
    dut = pipe.step()
    with pytest.raises(AlreadyDefinedError, match=r'another_designator.*some_designator'):
        dut.set_name('Some_Designator').set_name('Another_Designator')


def test_no_model() -> None:
    pipe = PipelineTemplate(name='foo', tags={})
    step = pipe.step()
    assert not step.has_model()


def test_has_model() -> None:
    pipe = PipelineTemplate(name='foo', tags={})
    step = pipe.step(model=ConnectorModel())
    assert step.has_model()


def test_args_override() -> None:
    pipe = PipelineTemplate(name='foo', tags={})
    step = pipe.step(model=ConnectorModel(), some_arg='some_value', another_arg='another_value')
    assert step.hyperparams(some_arg='different_value', third_arg=3) == {
        'some_arg': 'different_value',
        'another_arg': 'another_value',
        'third_arg': 3,
    }


def test_serialized_model() -> None:
    pipe = PipelineTemplate(name='foo')
    step = pipe.step(model=ConnectorModel(), serialized_model=b'pretend serialized model')
    assert step.serialized_model == b'pretend serialized model'


def test_filename() -> None:
    PipelineStep.reset_serial_number()
    pipe = PipelineTemplate(name='foo', tags={})
    unnamed_step = pipe.step()
    assert unnamed_step.filename is None

    model_step1 = pipe.step(model=ConnectorModel())
    model_step2 = pipe.step(model=ConnectorModel())

    assert model_step1.filename == '@connect_1@'
    assert model_step2.filename == '@connect_2@'

    named_step = pipe.step(model=ConnectorModel()).set_name('frob')
    assert named_step.filename == '@connect_3@frob'

    model_step4 = pipe.step(model=ConnectorModel())
    assert model_step4.filename == '@connect_4@'

    # Make sure that names are stable.
    assert model_step1.filename == '@connect_1@'



import numpy as np
import scipy.optimize as sp_opt  # type: ignore[import]

from .mallows_model import check_theta_phi, phi_to_theta

# ******** Complete rankings ********** #
# ************************************* #


def inverse(s):
    """ This function computes the inverse of a given permutation
        Parameters
        ----------
        s: ndarray
            A permutation array
        Returns
        -------
        ndarray
            The inverse of given permutation
    """
    return np.argsort(s)


def compose(s, p):
    """This function composes two given permutations
    Parameters
    ----------
    s: ndarray
        First permutation array
    p: ndarray
        Second permutation array
    Returns
    -------
    ndarray
        The composition of the permutations
    """
    return np.array(s[p])

# ************* Distance ************** #


def merge(left, right):
    """
    This function uses Merge sort algorithm to count the number of
    inversions in a permutation of two parts (left, right).
    Parameters
    ----------
    left: ndarray
        The first part of the permutation
    right: ndarray
        The second part of the permutation
    Returns
    -------
    result: ndarray
        The sorted permutation of the two parts
    count: int
        The number of inversions in these two parts.
    """
    result = []
    count = 0
    i, j = 0, 0
    left_len = len(left)
    while i < left_len and j < len(right):
        if left[i] <= right[j]:
            result.append(left[i])
            i += 1
        else:
            result.append(right[j])
            count += left_len - i
            j += 1
    result += left[i:]
    result += right[j:]

    return result, count


def mergeSort_rec(lst):
    """
    This function splits recursively lst into sublists until sublist size is 1. Then, it calls the function merge()
    to merge all those sublists to a sorted list and compute the number of inversions used to get that sorted list.
    Finally, it returns the number of inversions in lst.
    Parameters
    ----------
    lst: ndarray
        The permutation
    Returns
    -------
    result: ndarray
        The sorted permutation
    d: int
        The number of inversions.
    """
    lst = list(lst)
    if len(lst) <= 1:
        return lst, 0
    middle = int(len(lst) / 2)
    left, a = mergeSort_rec(lst[:middle])
    right, b = mergeSort_rec(lst[middle:])
    sorted_, c = merge(left, right)
    d = (a + b + c)
    return sorted_, d


def distance(A, B=None):
    """
    This function computes Kendall's-tau distance between two permutations
    using Merge sort algorithm.
    If only one permutation is given, the distance will be computed with the
    identity permutation as the second permutation
   Parameters
   ----------
   A: ndarray
        The first permutation
   B: ndarray, optional
        The second permutation (default is None)
   Returns
   -------
   int
        Kendall's-tau distance between both permutations (equal to the number of inversions in their composition).
    """
    if B is None:
        B = list(range(len(A)))

    A = np.asarray(A).copy()
    B = np.asarray(B).copy()
    n = len(A)

    # check if A contains NaNs
    msk = np.isnan(A)
    indexes = np.array(range(n))[msk]

    if indexes.size:
        A[indexes] = n  # np.nanmax(A)+1

    # check if B contains NaNs
    msk = np.isnan(B)
    indexes = np.array(range(n))[msk]

    if indexes.size:
        B[indexes] = n  # np.nanmax(B)+1

    # print(A,B,n)
    inverse = np.argsort(B)
    compose = A[inverse]
    _, distance = mergeSort_rec(compose)
    return distance


def max_dist(n):
    """ This function computes the maximum distance between two permutations of n length.
        Parameters
        ----------
        n: int
            Length of permutations
        Returns
        -------
        int
            Maximum distance between permutations of given n length.
    """
    return int(n * (n - 1) / 2)


# ************ Vector/Rankings ************** #


def v_to_ranking(v, n):
    """This function computes the corresponding permutation given a decomposition vector.
    The O(n log n) version in 10.1.1 of
    Arndt, J. (2010). Matters Computational: ideas, algorithms, source code.
    Springer Science & Business Media.
        Parameters
        ----------
        v: ndarray
            Decomposition vector, same length as the permutation, last item must be 0
        n: int
            Length of the permutation
        Returns
        -------
        ndarray
            The permutation corresponding to the decomposition vectors.
    """
    rem = list(range(n))
    rank = np.full(n, np.nan)
    for i in range(len(v)):
        rank[i] = rem[v[i]]
        rem.pop(v[i])
    return rank.astype(int)


def ranking_to_v(sigma, k=None):
    """This function computes the corresponding decomposition vector given a permutation
    The O(n log n) version in 10.1.1 of
    Arndt, J. (2010). Matters Computational: ideas, algorithms, source code.
    Springer Science & Business Media.
        Parameters
        ----------
        sigma: ndarray
            A permutation
        k: int, optional
            The index to perform the conversion for a partial
            top-k list
        Returns
        -------
        ndarray
            The decomposition vector corresponding to the permutation. Will be
            of length n and finish with 0.
    """
    n = len(sigma)
    if k is not None:
        sigma = sigma[:k]
        sigma = np.concatenate(
            (sigma, np.array([np.float(i) for i in range(n)
                              if i not in sigma])))
    V = []
    for j, sigma_j in enumerate(sigma):
        V_j = 0
        for i in range(j + 1, n):
            if sigma_j > sigma[i]:
                V_j += 1
        V.append(V_j)
    return np.array(V)


# ************ Sampling ************ #


def sample(m, n, *, k=None, theta=None, phi=None, s0=None):
    """This function generates m (rankings) according to Mallows Models (if the given parameters
    are m, n, k/None, theta/phi: float, s0/None) or Generalized Mallows Models (if the given
    parameters are m, n, theta/phi: ndarray, s0/None). Moreover, the parameter k allows the
    function to generate top-k rankings only.
        Parameters
        ----------
        m: int
            Number of rankings to generate
        n: int
            Length of rankings
        theta: float or ndarray, optional (if phi given)
            The dispersion parameter theta
        phi: float or ndarray, optional (if theta given)
            Dispersion parameter phi
        k: int
            Length of partial permutations (only top items)
        s0: ndarray
            Consensus ranking
        Returns
        -------
        ndarray
            The rankings generated
    """

    theta, phi = check_theta_phi(theta, phi)

    theta = np.full(n - 1, theta)

    if s0 is None:
        s0 = np.array(range(n))

    rnge = np.array(range(n - 1))

    psi = (1 - np.exp(
        (-n + rnge) * (theta[rnge]))) / (1 - np.exp(-theta[rnge]))
    vprobs = np.zeros((n, n))
    for j in range(n - 1):
        vprobs[j][0] = 1.0 / psi[j]
        for r in range(1, n - j):
            vprobs[j][r] = np.exp(-theta[j] * r) / psi[j]
    sample = []
    for samp in range(m):
        v = [np.random.choice(n, p=vprobs[i, :]) for i in range(n - 1)]
        v += [0]
        ranking = v_to_ranking(v, n)
        sample.append(ranking)

    sample = np.array([s[s0] for s in sample])

    if k is not None:
        sample_rankings = np.array(
            [inverse(ordering) for ordering in sample])
        sample_rankings = np.array([ran[s0] for ran in sample_rankings])
        sample = np.array([[i if i in range(k) else np.nan for i in ranking]
                           for ranking in sample_rankings])
    return sample


def num_perms_at_dist(n):
    """This function computes the number of permutations of length 1 to n for
    each possible Kendall's-tau distance d. See the online Encyclopedia of
    Integer Sequences, OEIS-A008302
        Parameters
        ----------
        n: int
            Length of the permutations
        Returns
        -------
        ndarray
            The number of permutations of length 1 to n for each possible
            Kendall's-tau distance d
    """
    sk = np.zeros((n + 1, int(n * (n - 1) / 2 + 1)))
    for i in range(n + 1):
        sk[i, 0] = 1
    for i in range(1, 1 + n):
        for j in range(1, int(i * (i - 1) / 2 + 1)):
            if j - i >= 0:
                sk[i, j] = sk[i, j - 1] + sk[i - 1, j] - sk[i - 1, j - i]
            else:
                sk[i, j] = sk[i, j - 1] + sk[i - 1, j]
    return sk.astype(np.uint64)


def sample_at_dist(n, dist, sk=None, sigma0=None):
    """This function randomly generates a permutation with length n at distance
    dist to a given permutation sigma0.
        Parameters
        ----------
        n: int
            Length of the permutations
        dist: int
            Distance between the permutation generated randomly and a known
            permutation sigma0
        sk: matrix
            matrix returned by the function mallows_kendall::num_perms_at_dist(n)
            if this function is to be called many times, to avoid recomputation,
            sk can be provided in the input. Otherwise, the function is called here
        sigma0: ndarray, optional
            A known permutation (If not given, then it equals the identity)
        Returns
        -------
        ndarray
            A random permutation at distance dist to sigma0.
    """
    i = 0
    probs = np.zeros(n + 1)
    v = np.zeros(n, dtype=int)
    if sk is None:
        sk = num_perms_at_dist(n)

    while i < n and dist > 0:
        rest_max_dist = (n - i - 1) * (n - i - 2) / 2
        if rest_max_dist >= dist:
            probs[0] = sk[n - i - 1, dist]
        else:
            probs[0] = 0
        mi = min(dist + 1, n - i)
        for j in range(1, mi):
            if rest_max_dist + j >= dist:
                probs[j] = sk[n - i - 1, dist - j]
            else:
                probs[j] = 0
        v[i] = np.random.choice(mi, 1, p=probs[:mi] / probs[:mi].sum())
        dist -= v[i]
        i += 1
    random_perm = v_to_ranking(v, n)

    return random_perm[sigma0].reshape(-1)


# ********* Expected distance ********* #


def expected_dist_mm(n, theta=None, phi=None):
    """The function computes the expected distance of Kendall's-tau distance under Mallows models (MMs).
        Parameters
        ----------
        n: int
            Length of the permutation in the considered model
        theta: float
            Real dispersion parameter, optional (if phi is given)
        phi: float
            Real dispersion parameter, optional (if theta is given)
        Returns
        -------
        float
            The expected distance under MMs.
    """
    theta, phi = check_theta_phi(theta, phi)
    rnge = np.array(range(1, n + 1))
    expected_dist = n * np.exp(-theta) / (1 - np.exp(-theta)) - np.sum(
        rnge * np.exp(-rnge * theta) / (1 - np.exp(-rnge * theta)))

    return expected_dist


# ************ Variance ************ #


def variance_dist_mm(n, theta=None, phi=None):
    """This function returns the variance of Kendall's-tau distance under the MMs.

        Parameters
        ----------
        n: int
            Length of the permutations
        theta: float
            Dispersion parameter, optional (if phi is given)
        phi  : float
            Dispersion parameter, optional (if theta is given)
        Returns
        -------
        float
            The variance of Kendall's-tau distance under the MMs.
    """
    theta, phi = check_theta_phi(theta, phi)
    rnge = np.array(range(1, n + 1))
    variance = (phi * n) / (1 - phi)**2 - np.sum(
        (pow(phi, rnge) * rnge**2) / (1 - pow(phi, rnge))**2)

    return variance


# ************ Learning ************ #


def median(rankings):  # Borda
    """This function computes the central permutation (consensus ranking) given

    several permutations.

        Parameters
        ----------
        rankings: ndarray
            Matrix of several permutations
        Returns
        -------
        ndarray
            The central permutation of permutations given.
    """
    consensus = np.argsort(  # give the inverse of result --> sigma_0
        np.argsort(  # give the indexes to sort the sum vector --> sigma_0^-1
            rankings.sum(axis=0)  # sum the indexes of all permutations
        ))
    return consensus


def fit_mm(rankings, s0=None):
    """This function computes the consensus ranking and the MLE for the
    dispersion parameter phi for MM models.
        Parameters
        ----------
        rankings: ndarray
            The matrix of permutations
        s0: ndarray, optional
            The consensus ranking (default value is None)
        Returns
        -------
        tuple
            The ndarray corresponding to s0 the consensus permutation and the
            MLE for the dispersion parameter phi.
    """
    m, n = rankings.shape
    if s0 is None:
        s0 = np.argsort(np.argsort(rankings.sum(axis=0)))  # borda
    dist_avg = np.mean(np.array([distance(s0, perm) for perm in rankings]))
    try:
        theta = sp_opt.newton(mle_theta_mm_f,
                              0.01,
                              fprime=mle_theta_mm_fdev,
                              args=(n, dist_avg),
                              tol=1.48e-08,
                              maxiter=500,
                              fprime2=None)
    except Exception:
        if dist_avg == 0.0:
            return s0, np.exp(-5)  # =phi
        print("Error in function: fit_mm. dist_avg=", dist_avg,
              dist_avg == 0.0)
        print(rankings)
        print(s0)
        raise
    return s0, np.exp(-theta)  # =phi


# ************ Top-k rankings ************ #
# **************************************** #

# *************** Distance *************** #


def p_distance(beta_1, beta_2, k, p=0):
    """This function returns the distance between top-k rankings using
    the p-parametrized Kendall's-tau distance.
    Parameters
    ----------
    beta_1: ndarray
        A top-k permutation
    beta_2: ndarray
        A top-k permutation
    k: int
        Length of partial permutations (only top items)
    p: float
        The parameter in [0, 1]
    Returns
    -------
    float
        The p-parametrized Kendall's-tau distance.
    """

    alpha_1 = beta_to_alpha(beta_1, k=k)
    alpha_2 = beta_to_alpha(beta_2, k=k)
    d = 0
    p_counter = 0
    alpha_1Ualpha_2 = list(
        set(
            int(x) for x in np.union1d(alpha_1, alpha_2)
            if not np.isnan(x)))
    for i_index, i in enumerate(alpha_1Ualpha_2):
        i_1_nan = np.isnan(beta_1[i])
        i_2_nan = np.isnan(beta_2[i])
        for j in alpha_1Ualpha_2[i_index + 1:]:
            j_1_nan = np.isnan(beta_1[j])
            j_2_nan = np.isnan(beta_2[j])
            if not i_1_nan and not j_1_nan and not i_2_nan and not j_2_nan:
                if (beta_1[i] > beta_1[j] and beta_2[i] > beta_2[j]) or \
                   (beta_1[i] < beta_1[j] and beta_2[i] < beta_2[j]):
                    continue
                elif (beta_1[i] > beta_1[j] and beta_2[i] < beta_2[j]) or \
                     (beta_1[i] < beta_1[j] and beta_2[i] > beta_2[j]):
                    d += 1
            elif (not i_1_nan and not j_1_nan and ((not i_2_nan and j_2_nan) or (i_2_nan and not j_2_nan))) or \
                 (not i_2_nan and not j_2_nan and ((not i_1_nan and j_1_nan) or (i_1_nan and not j_1_nan))):
                if i_1_nan:
                    d += int(beta_2[j] > beta_2[i])
                elif j_1_nan:
                    d += int(beta_2[i] > beta_2[j])
                elif i_2_nan:
                    d += int(beta_1[j] > beta_1[i])
                elif j_2_nan:
                    d += int(beta_1[i] > beta_1[j])
            elif (not i_1_nan and j_1_nan and i_2_nan and not j_2_nan) or \
                 (i_1_nan and not j_1_nan and not i_2_nan and j_2_nan):
                d += 1
            elif (not i_1_nan and not j_1_nan and i_2_nan and j_2_nan) or \
                 (i_1_nan and j_1_nan and not i_2_nan and not j_2_nan):
                p_counter += 1
    return d + p_counter * p


# ********** Expected distance ********** #


def expected_dist_top_k(n, k, theta=None, phi=None):
    """Compute the expected distance for top-k rankings, following
    a MM under the Kendall's-tau distance.
        Parameters
        ----------
        n: int
            Length of the permutation in the considered model
        k: int
            Length of partial permutations (only top items)
        theta: float, optional (if phi is given)
            Real dispersion parameter
        phi  : float, optional (if theta is given)
            Real dispersion parameter
        Returns
        -------
        float
            The expected disance under the MMs.
    """
    theta, phi = check_theta_phi(theta, phi)
    rnge = np.array(range(n - k + 1, n + 1))
    expected_dist = (k * phi / (1 - phi)
                     - np.sum(rnge * pow(phi, rnge) / (1 - pow(phi, rnge))))
    return expected_dist


# ************ Variance ************* #


def variance_dist_top_k(n, k, theta=None, phi=None):
    """Compute the variance of the distance for top-k rankings, following
        a MM under the Kendall's-tau distance.
        Parameters
        ----------
        n: int
            Length of the permutation in the considered model
        k: int
            Length of partial permutations (only top items)
        theta: float, optional (if phi is given)
            Real dispersion parameter
        phi  : float, optional (if theta is given)
            Real dispersion parameter
        Returns
        -------
        float
            The variance under the MMs.
    """
    theta, phi = check_theta_phi(theta, phi)
    rnge = np.array(range(n - k + 1, n + 1))
    variance = (phi * k) / (1 - phi)**2 - np.sum(
        (pow(phi, rnge) * rnge**2) / (1 - pow(phi, rnge))**2)
    return variance


# ***** Expected/Variance vector ***** #


def expected_v(n, theta=None, phi=None, k=None):  # txapu integrar
    """This function computes the expected decomposition vector.
        Parameters
        ----------
        n: int
            Length of the permutation in the considered model
        theta: float, optional (if phi is given)
            Real dispersion parameter
        phi  : float, optional (if theta is given)
            Real dispersion parameter
        k: int, optional
            Length of partial permutations (only top items)
        Returns
        -------
        ndarray
            The expected decomposition vector.
    """
    theta, phi = check_theta_phi(theta, phi)
    if k is None:
        k = n - 1
    if not isinstance(theta, list):
        theta = np.full(k, theta)
    rnge = np.array(range(k))
    expected_v = np.exp(-theta[rnge]) / (1 - np.exp(-theta[rnge])) - (
        n - rnge) * np.exp(-(n - rnge) * theta[rnge]) / (
            1 - np.exp(-(n - rnge) * theta[rnge]))
    return expected_v


def variance_v(n, theta=None, phi=None, k=None):
    """This function computes the variance of the decomposition vector under GMM.
        Parameters
        ----------
        n: int
            Length of the permutation in the considered model
        theta: float, optional (if phi is given)
            Real dispersion parameter
        phi  : float, optional (if theta is given)
            Real dispersion parameter
        k: int, optional
            Length of partial permutations (only top items)
        Returns
        -------
        ndarray
            The variance of the decomposition vector.
    """
    theta, phi = check_theta_phi(theta, phi)
    if k is None:
        k = n - 1
    if not isinstance(phi, list):
        phi = np.full(k, phi)
    rnge = np.array(range(k))
    var_v = phi[rnge] / (1 - phi[rnge])**2 - (n - rnge)**2 * phi[rnge]**(
        n - rnge) / (1 - phi[rnge]**(n - rnge))**2
    return var_v


# ******** More functions *********#


def prob(sigma, sigma0, theta=None, phi=None):
    """Probability mass function of a MM with central ranking sigma0 and
    dispersion parameter theta/phi.
    Parameters
    ----------
    sigma: ndarray
        A pemutation
    sigma0: ndarray
        central permutation
    theta: float
        Dispersion parameter (optional, if phi is given)
    phi: float
        Dispersion parameter (optional, if theta is given)
    Returns
    -------
    float
        Probability mass function.
    """
    n = len(sigma)
    theta, phi = check_theta_phi(theta, phi)
    sigma0_inv = inverse(sigma0)
    rnge = np.array(range(n - 1))

    psi = (1 - np.exp((-n + rnge) * (theta))) / (1 - np.exp(-theta))
    psi = np.prod(psi)

    dist = distance(compose(sigma, sigma0_inv))

    return np.exp(-theta * dist) / psi


def borda_partial(rankings, w, k):
    """This function approximate the consensus ranking of a top-k rankings using Borda algorithm.
        Each nan-ranked item is assumed to have ranking $k$
        Parameters
        ----------
        rankings: ndarray
            The matrix of permutations
        w: float
            weight of each ranking
        k: int
            Length of partial permutations (only top items)
        Returns
        -------
        ndarray
            Consensus ranking.
    """
    a, b = rankings, w
    a, b = np.nan_to_num(rankings, nan=k), w
    aux = a * b
    borda = np.argsort(np.argsort(np.nanmean(aux, axis=0))).astype(float)
    mask = np.isnan(rankings).all(axis=0)
    borda[mask] = np.nan
    return borda


def psi_mm(n, theta=None, phi=None):
    """This function computes the normalization constant psi.
        Parameters
        ----------
        n: int
            Length of the permutation in the considered model
        theta: float
            Real dispersion parameter (optional if phi is given)
        phi: float
            Real dispersion parameter (optional if theta is given)
        Returns
        -------
        float
            The normalization constant psi.
    """
    rnge = np.array(range(2, n + 1))
    if theta is not None:
        return np.prod((1 - np.exp(-theta * rnge)) / (1 - np.exp(-theta)))
    if phi is not None:
        return np.prod((1 - np.power(phi, rnge)) / (1 - phi))
    theta, phi = check_theta_phi(theta, phi)


def fit_gmm(rankings, s0=None):
    """This function computes the consensus permutation and the MLE for the
    dispersion parameters theta_j for GMM models
        Parameters
        ----------
        rankings: ndarray
            The matrix of permutations
        s0: ndarray, optional
            The consensus permutation (default value is None)
        Returns
        -------
        tuple
            The ndarray corresponding to s0 the consensus permutation and the
            MLE for the dispersion parameters theta
    """
    m, n = rankings.shape
    if s0 is None:
        s0 = np.argsort(np.argsort(rankings.sum(axis=0)))  # borda
    V_avg = np.mean(np.array([ranking_to_v(sigma)[:-1] for sigma in rankings]),
                    axis=0)
    try:
        theta = []
        for j in range(1, n):
            theta_j = sp_opt.newton(mle_theta_j_gmm_f,
                                    0.01,
                                    fprime=mle_theta_j_gmm_fdev,
                                    args=(n, j, V_avg[j - 1]),
                                    tol=1.48e-08,
                                    maxiter=500,
                                    fprime2=None)
            theta.append(theta_j)
    except Exception:
        print("Error in function fit_gmm")
        raise
    return s0, theta


def mle_theta_mm_f(theta, n, dist_avg):
    """Compute the derivative of the likelihood.
    parameter
        Parameters
        ----------
        theta: float
            The dispersion parameter
        n: int
            Length of the permutations
        dist_avg: float
            Average distance of the sample (between the consensus and the
            permutations of the consensus)
        Returns
        -------
        float
            Value of the function for given parameters.
    """
    aux = 0
    rnge = np.array(range(1, n))
    aux = np.sum((n - rnge + 1) * np.exp(-theta * (n - rnge + 1))
                 / (1 - np.exp(-theta * (n - rnge + 1))))
    aux2 = (n - 1) / (np.exp(theta) - 1) - dist_avg

    return aux2 - aux


def mle_theta_mm_fdev(theta, n, dist_avg):
    """This function computes the derivative of the function mle_theta_mm_f
    given the dispersion parameter and the average distance.
        Parameters
        ----------
        theta: float
            The dispersion parameter
        n: int
            Length of the permutations
        dist_avg: float
            Average distance of the sample (between the consensus and the
            permutations of the consensus)
        Returns
        -------
        float
            The value of the derivative of function mle_theta_mm_f for given
            parameters.
    """
    aux = 0
    rnge = np.array(range(1, n))
    aux = np.sum((n - rnge + 1) * (n - rnge + 1)
                 * np.exp(-theta * (n - rnge + 1)) / pow(
                     (1 - np.exp(-theta * (n - rnge + 1))), 2))
    aux2 = (-n + 1) * np.exp(theta) / pow((np.exp(theta) - 1), 2)

    return aux2 + aux


def mle_theta_j_gmm_f(theta_j, n, j, v_j_avg):
    """Compute the derivative of the likelihood parameter theta_j in the GMM.
        Parameters
        ----------
        theta: float
            The jth dispersion parameter theta_j
        n: int
            Length of the permutations
        j: int
            The position of the theta_j in vector theta of dispersion parameters
        v_j_avg: float
            jth element of the average decomposition vector over the sample
        Returns
        -------
        float
            Value of the function for given parameters.
    """
    f_1 = np.exp(-theta_j) / (1 - np.exp(-theta_j))
    f_2 = (-(n - j + 1)
           * np.exp(-theta_j * (n - j + 1))
           / (1 - np.exp(-theta_j * (n - j + 1))))
    return f_1 + f_2 - v_j_avg


def mle_theta_j_gmm_fdev(theta_j, n, j, v_j_avg):
    """This function computes the derivative of the function mle_theta_j_gmm_f
    given the jth element of the dispersion parameter and the jth element of the
    average decomposition vector.
        Parameters
        ----------
        theta: float
            The jth dispersion parameter theta_j
        n: int
            Length of the permutations
        j: int
            The position of the theta_j in vector theta of dispersion parameters
        v_j_avg: float
            jth element of the average decomposition vector over the sample
        Returns
        -------
        float
            The value of the derivative of function mle_theta_j_gmm_f for given
            parameters.
    """
    fdev_1 = -np.exp(-theta_j) / pow((1 - np.exp(-theta_j)), 2)
    fdev_2 = pow(n - j + 1, 2) * np.exp(-theta_j * (n - j + 1)) / pow(
        1 - np.exp(-theta_j * (n - j + 1)), 2)
    return fdev_1 + fdev_2


def likelihood_mm(perms, s0, theta):
    """This function computes the log-likelihood for MM model given a matrix of
    permutation, the consensus permutation, and the dispersion parameter.
        Parameters
        ----------
        perms: ndarray
            A matrix of permutations
        s0: ndarray
            The consensus permutation
        theta: float
            The dispersion parameter
        Returns
        -------
        float
            Value of log-likelihood for given parameters.
    """
    m, n = perms.shape
    rnge = np.array(range(2, n + 1))
    psi = 1.0 / np.prod((1 - np.exp(-theta * rnge)) / (1 - np.exp(-theta)))
    probs = np.array(
        [np.log(np.exp(-distance(s0, perm) * theta) / psi) for perm in perms])

    return probs.sum()


def alpha_to_beta(alpha, k):  # aux for the p_distance
    inv = np.full(len(alpha), np.nan)
    for i, j in enumerate(alpha[:k]):
        inv[int(j)] = i
    return inv


def beta_to_alpha(beta, k):  # aux for the p_distance
    inv = np.full(len(beta), np.nan)
    for i, j in enumerate(beta):
        if not np.isnan(j):
            inv[int(j)] = i
    return inv


def find_phi_n(n, bins):
    """ Divide the expected distances into bins and return both the expected
    distances and their corresponding values of dispersion parameter phi.
    Parameters
    ----------
    n: int
        Length of permutations
    bins: int
        Number of bins
    Returns
    -------
    tuple
        An array of expected distances and their corresponding dispersion parameter phi
    """
    ed, phi_ed = [], []
    ed_uniform = (n * (n - 1) / 2) / 2
    for dmin in np.linspace(0, ed_uniform - 1, bins):
        ed.append(dmin)
        phi_ed.append(find_phi(n, dmin, dmin + 1))
    return ed, phi_ed


def find_phi(n, dmin, dmax):
    """Find the dispersion parameter phi that gives an expected distance between
    dmin and dmax where the length of rankings is n.
    Parameters
    ----------
    n: int
        Length of permutations
    dmin: int
        The minimum of expected distance
    dmax: int
        The maximum of expected distance
    Returns
    -------
    float
        The value of phi.
    """
    imin, imax = np.float64(0), np.float64(1)
    iterat = 0
    while iterat < 500:
        med = imin + (imax - imin) / 2
        d = expected_dist_mm(n, phi_to_theta(med))
        if d < dmax and d > dmin:
            return med
        elif d < dmin:
            imin = med
        elif d > dmax:
            imax = med
        iterat += 1


# end



'''Tests for SimpleStep'''
# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import re

from ...algorithms.connect import ConnectorModel
from ...algorithms.impl.fake_algorithm import FakeAlgorithm, FAKE_SERIALIZED
from ...generator.bound_pipeline import BoundPipeline
from ...generator.designator import StepDesignator
from ...templates.impl.pipeline_template import PipelineTemplate
from ...wrangler.dataset import Dataset
from .simple_executable_step import SimpleExecutableStep


# pylint: disable=missing-function-docstring,missing-class-docstring,duplicate-code,protected-access

REFERENCE_DATASET = Dataset(
    a_key='a_value',
    another_key='another_value',
    ignored_key='ignored_value'
)


def strip_serial(key: StepDesignator) -> str:
    return re.sub('[0-9]+', 'NUM', key)


def test_sunny_day() -> None:
    pipe = PipelineTemplate(name='foo', tags={})
    model = ConnectorModel()
    step = pipe.step(model=model, new_key='a_key', another_key='another_key')
    dut = SimpleExecutableStep(bound_step=step)
    got = dut.predict(dataset=REFERENCE_DATASET)
    want = {
        'new_key': 'a_value',
        'another_key': 'another_value',
    }
    assert got == want


def test_name_propagation() -> None:
    '''Test that when you create an executable pipeline from a bound pipeline,
    the name and serial number propagates correctly.
    '''
    pipe = PipelineTemplate(name='foo', tags={})
    model = ConnectorModel()
    step = pipe.step(model=model).set_name('the_name_we_want')
    dut = SimpleExecutableStep(bound_step=step)
    assert dut.opt_name == 'the_name_we_want'
    assert dut.filename == step.filename
    assert dut.filename is not None
    assert strip_serial(dut.filename) == '@connect_NUM@the_name_we_want'


def test_cucumberize_all() -> None:
    pipe = BoundPipeline(name='foo', tags={})
    model = ConnectorModel(some_param='some_value')
    step = pipe.step(model=model,
                     bound_added_param='bound_added_value').set_name('the_name_we_want')
    dut = SimpleExecutableStep(
        bound_step=step)

    got_jar = dut.cucumberize_all(pipeline_designator=pipe.designator)
    assert len(got_jar) == 1
    got_cucumber = list(got_jar.values())[0]
    assert strip_serial(got_cucumber.filename) == '@connect_NUM@the_name_we_want'
    assert got_cucumber.catalog_name == 'connect'
    assert got_cucumber.pipeline_designator == pipe.designator
    assert got_cucumber.hyperparams == {
        'some_param': 'some_value',
        'bound_added_param': 'bound_added_value'}


def test_serialize_deserialize_instance():
    pipe = BoundPipeline(name='foo', tags={})
    model = FakeAlgorithm(some_param='some_value')
    instance = model.instantiate()
    serialized_model = instance.serialize()
    assert instance.evidence_of_deserialize is None
    step = pipe.step(model=model,
                     serialized_model=serialized_model)
    dut = SimpleExecutableStep(bound_step=step)
    assert dut._model_instance.evidence_of_deserialize == FAKE_SERIALIZED



'''Tests for wrangler.py'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=missing-function-docstring,duplicate-code,protected-access,too-many-ancestors
# pylint: disable=missing-class-docstring,redefined-outer-name,unused-variable

import glob
import json
import os
from pathlib import Path
from typing import Dict, Optional, Tuple

from numpy.random import RandomState  # pylint: disable=no-name-in-module  # This just isn't true.
import pandas as pd
import pytest

from ..aggregators.impl.aggregator import AggregatorStub
from ..algorithms.impl.algorithm import Algorithm
from ..algorithms.impl.algorithm_auto import FakeAlgorithmCatalogAuto
from ..algorithms.impl.algorithm_instance import AlgorithmInstance
from ..algorithms.impl.fake_algorithm import FakeInstance
from ..executor.cucumber import Cucumber
from ..executor.executor import ExecutorStub
from ..executor.simple.simple_executor import SimpleExecutor
from ..generator.designator import Designator
from ..instantiator.executable_pipeline import (ExecutablePipeline,
                                                PipelineResult)
from ..problem_def.problem_def import ProblemDefinition
from ..templates.impl.pipeline_step import PipelineStep
from ..templates.impl.pipeline_template import PipelineTemplate
from ..templates.impl.template import TemplateCatalog

from .constants import Defaults
from .dataset import Dataset, RoleName
from .wrangler import Wrangler


class FakeExecutor(ExecutorStub):

    def predict(self,
                dataset: Dataset,
                pipelines: Dict[Designator, ExecutablePipeline]
                ) -> Dict[Designator, PipelineResult]:
        '''Use a list of pipelines to predict from a dataset.'''
        predictions = dataset.output()
        # create a dataframe with the shape we expect predictions to be:
        # 1 column (the target) and a number of rows equal to the size of the input
        predictions.predictions = pd.DataFrame({
            dataset.metadata.roles[RoleName.TARGET][0].name:
                range(0, dataset.dataframe.shape[0])})
        return {d: PipelineResult(executable_pipeline=p, prediction=predictions)
                for d, p in pipelines.items()}


MEMORY_PROBLEM_DEF = '''
{
    "dataset": {
        "config": "memory",
        "column_roles": {
            "target": {
                "name": "c"
            }
        },
        "params": {
            "train_data": "train_data",
            "test_data": "test_data"
        }
    },
    "problem_type": {
        "task": "regression"
    },
    "cross_validation": {
        "k": 10
    },
    "metrics" : {
        "root_mean_squared_error" : {}
    },
    "hyperparams": ["disable_grid_search"]
}
'''

AGGREGATION_PROBLEM_DEF = '''
{
    "dataset": {
        "config": "memory",
        "column_roles": {
            "target": {
                "name": "c"
            }
        },
        "params": {
            "train_data": "train_data",
            "test_data": "test_data"
        }
    },
    "problem_type": {
        "task": "regression"
    },
    "cross_validation": {
        "k": 10
    },
    "metrics": {
        "root_mean_squared_error" : {},
        "mean_squared_error": {},
        "mean_absolute_error": {}
    },
    "aggregation": {
        "method": ["method1", "method2"]
    },
    "hyperparams": ["disable_grid_search"]
}
'''


def get_data() -> Tuple[pd.DataFrame, Optional[pd.DataFrame]]:
    data = pd.DataFrame(
        {
            'a': range(1, 21),
            'b': range(1, 21),
            'c': range(1, 21)
        }
    )
    return (data, data)


def memory_problem_def(has_aggregation: bool = False) -> ProblemDefinition:
    problem_def = MEMORY_PROBLEM_DEF
    if has_aggregation:
        problem_def = AGGREGATION_PROBLEM_DEF

    return ProblemDefinition(clause=problem_def)


def test_wrangler_split() -> None:
    (train_data, test_data) = get_data()  # noqa: F841
    problem_def = memory_problem_def()
    dut = Wrangler(
        problem_definition=problem_def,
        algorithm_catalog=FakeAlgorithmCatalogAuto,
        executor=FakeExecutor)

    train_result = dut.fit_predict_rank()

    # we expect 10 folds because k=10 was set in the problem definition
    assert len(train_result.split_dataset.folds) == 10
    # we expect each fold's validate set to have 2 rows,
    #    because there are 10 folds and the dataset is 20 rows long,
    #    and 2 columns, because it contains the attributes but not the target.
    assert train_result.split_dataset.folds[0].validate.dataframe.shape == (2, 2)
    # we expect each fold's train set to have 18 rows,
    #    because it is composed of 9 folds of 2 rows each,
    #    and 3 columns, becase it contains the attributes and the target.
    assert train_result.split_dataset.folds[0].train.dataframe.shape == (18, 3)
    # we expect the ground truth to have 20 rows,
    #    because it covers the whole dataset,
    #    and 1 col, because it only has the target.
    assert train_result.split_dataset.ground_truth is not None
    assert train_result.split_dataset.ground_truth.ground_truth.shape == (20, 1)


def test_fit_predict_rank() -> None:
    (train_data, test_data) = get_data()  # noqa: F841
    problem_def = memory_problem_def()

    dut = Wrangler(
        problem_definition=problem_def,
        algorithm_catalog=FakeAlgorithmCatalogAuto)

    got = dut.fit_predict_rank()

    # we expect pipelines to be named according to thier template (tabular regression)
    #    and the model that was chosen from thier query step.
    #    in the case of random forest regressor, wrangler will do grid search,
    #    and thus hyperparam values will also be in the designator.
    random_forest_des = Designator('tabular_regression@sklearn.ensemble.randomforestregressor:'
                                   'max_depth=none,min_samples_split=2')
    linear_regression_des = Designator('tabular_regression@sklearn.linear_model.linearregression')
    assert {random_forest_des, linear_regression_des} == set(got.executable_pipelines.keys())

    assert len(got.train_results) == 2  # 2 pipelines
    got_result = got.train_results.predictions[random_forest_des]
    assert got_result is not None
    assert got_result.predictions is not None
    assert got_result.predictions.shape == (20, 2)

    assert len(got.rankings) == 1  # 1 metric
    assert 'root_mean_squared_error' in got.rankings

    assert got.test_results is not None
    assert len(got.test_results) == 2  # 2 pipelines

    # executable pipelines associated with the test results should be identical
    assert set(got.executable_pipelines.keys()) == set(got.test_results.executable_pipelines.keys())
    got_result = got.test_results.predictions[random_forest_des]
    assert got_result is not None
    assert got_result.predictions is not None
    assert got_result.predictions.shape == (20, 1)


def test_fit_predict_rank_with_aggregation() -> None:
    (train_data, test_data) = get_data()  # noqa: F841
    problem_def = memory_problem_def(has_aggregation=True)

    dut = Wrangler(
        problem_definition=problem_def,
        algorithm_catalog=FakeAlgorithmCatalogAuto)

    dut.aggregator_catalog.register(AggregatorStub(name="method1"))
    dut.aggregator_catalog.register(AggregatorStub(name="method2"))

    got = dut.fit_predict_rank()

    assert len(got.rankings) == 5  # 3 metrics, and 2 aggregate metrics
    want = {
        'root_mean_squared_error', 'mean_squared_error', 'mean_absolute_error',
        'method1', 'method2'
    }
    assert want == set(got.rankings.keys())


class FakeTemplateCatalog(TemplateCatalog):
    '''fake, no templates'''


def memory_test_df() -> pd.DataFrame:
    return pd.DataFrame(
        {
            'a': range(1, 11),
            'b': range(1, 11)
        }
    )


def memory_train_df() -> pd.DataFrame:
    return pd.DataFrame(
        {
            'a': range(1, 21),
            'b': range(1, 21),
            'c': range(1, 21)
        }
    )


def test_predict() -> None:
    (train_data, test_data) = get_data()  # noqa: F841
    problem_def = memory_problem_def()
    executor = SimpleExecutor
    alg_catalog = FakeAlgorithmCatalogAuto
    dut = Wrangler(
        problem_definition=problem_def,
        executor=executor,
        algorithm_catalog=alg_catalog)
    got = dut.fit_predict_rank()

    fake_des = Designator('fake_des')

    testset = dut.dataset(memory_test_df())
    got_result = list(got.train_results.values())[0]
    pipeline = got_result.executable_pipeline
    assert pipeline is not None
    test_got = dut.predict(
        new_data=testset,
        trained_pipelines={fake_des: pipeline})
    got_result = test_got[fake_des]
    assert got_result is not None
    assert got_result.prediction is not None
    prediction_df = got_result.prediction.predictions

    # we expect 10 rows (length of test dataframe) and 1 col (target)
    assert prediction_df.shape == (10, 1)
    # we expect to be able to predict perfectly since test data is identical to train data.
    assert prediction_df['c'][9] == 10.0


def test_predict_defaulted() -> None:
    (train_data, test_data) = get_data()  # noqa: F841
    problem_def = memory_problem_def()
    executor = SimpleExecutor
    alg_catalog = FakeAlgorithmCatalogAuto
    dut = Wrangler(
        problem_definition=problem_def,
        executor=executor,
        algorithm_catalog=alg_catalog)
    _ = dut.fit_predict_rank()

    testset = dut.dataset(memory_test_df())
    test_got = dut.predict(
        new_data=testset)
    assert {
        'tabular_regression@sklearn.linear_model.linearregression',
        'tabular_regression@sklearn.ensemble.randomforestregressor:'
        'max_depth=none,min_samples_split=2',
    } == set(test_got.keys())


def test_set_current_predict() -> None:
    (train_data, test_data) = get_data()  # noqa: F841
    problem_def = memory_problem_def()
    executor = SimpleExecutor
    alg_catalog = FakeAlgorithmCatalogAuto
    dut = Wrangler(
        problem_definition=problem_def,
        executor=executor,
        algorithm_catalog=alg_catalog)
    _ = dut.fit_predict_rank()

    testset = dut.dataset(memory_test_df())

    # Pick one pipeline.
    dut.set_current('tabular_regression@sklearn.ensemble.randomforestregressor:'
                    'max_depth=none,min_samples_split=2')
    test_got = dut.predict(new_data=testset)
    assert set(test_got.keys()) == {
        'tabular_regression@sklearn.ensemble.randomforestregressor:'
        'max_depth=none,min_samples_split=2',
    }

    # Reset to all known pipelines.
    dut.set_current()

    test_got = dut.predict(new_data=testset)
    assert {
        'tabular_regression@sklearn.linear_model.linearregression',
        'tabular_regression@sklearn.ensemble.randomforestregressor:'
        'max_depth=none,min_samples_split=2',
    } == set(test_got.keys())


def test_set_current_fit() -> None:
    (train_data, test_data) = get_data()
    problem_def = memory_problem_def()
    executor = SimpleExecutor
    alg_catalog = FakeAlgorithmCatalogAuto
    dut = Wrangler(
        problem_definition=problem_def,
        executor=executor,
        algorithm_catalog=alg_catalog)
    _ = dut.fit_predict_rank()

    testset = dut.dataset(memory_train_df())

    # Pick one pipeline.
    dut.set_current('tabular_regression@sklearn.ensemble.randomforestregressor:'
                    'max_depth=none,min_samples_split=2')

    test_got = dut.fit(dataset=testset)
    assert set(test_got.keys()) == {
        'tabular_regression@sklearn.ensemble.randomforestregressor:'
        'max_depth=none,min_samples_split=2',
    }

    # Reset to all known pipelines.
    dut.set_current()

    test_got = dut.fit(dataset=testset)
    assert {
        'tabular_regression@sklearn.linear_model.linearregression',
        'tabular_regression@sklearn.ensemble.randomforestregressor:'
        'max_depth=none,min_samples_split=2',
    } == set(test_got.keys())


@pytest.fixture(scope='session')
def saver_problem_def(tmp_path_factory: pytest.TempPathFactory) -> ProblemDefinition:
    tmp_path = tmp_path_factory.mktemp('data')
    tmp_dir = tmp_path / 'sub'
    tmp_dir.mkdir()
    config = {
        "dataset": {
            "config": "memory",
            "column_roles": {
                "target": {
                    "name": "c"
                }
            },
            "params": {
                "train_data": "train_data",
                "test_data": "test_data"
            }
        },
        "problem_type": {
            "task": "regression"
        },
        "cross_validation": {
            "k": 10
        },
        "metrics": {
            "root_mean_squared_error": {}
        },
        "output": {
            "path": tmp_dir,
            "instantiations": [
                "json"
            ],
            "file_type": "csv"
        },
        "hyperparams": ["disable_grid_search"]
    }

    return ProblemDefinition(clause=config)


def test_save_filesystem(saver_problem_def: ProblemDefinition) -> None:
    PipelineStep.reset_serial_number()
    (train_data, test_data) = get_data()
    dut = Wrangler(
        problem_definition=saver_problem_def,
        algorithm_catalog=FakeAlgorithmCatalogAuto)
    _ = dut.fit_predict_rank()

    path = saver_problem_def.output.path
    assert path is not None
    filenames = glob.glob(str(path / 'models' / '*.linearregression' / '@connect_*@.pkl'))
    assert len(filenames) == 1
    with Path(filenames[0]).open(mode='rb') as filepointer:
        cucumber = Cucumber.deserialize(filepointer.read())
    assert cucumber.hyperparams == {
        'target': ['target', 'dataframe'],
        'covariates': ['attributes', 'dataframe']
    }

    pipelines = os.listdir(str(path / 'pipelines'))
    assert {'tabular_regression@sklearn.ensemble.randomforestregressor:'
            'max_depth=none,min_samples_split=2.json',
            'tabular_regression@sklearn.linear_model.linearregression.json',
            } == set(pipelines)

    with (path / 'pipelines' / 'tabular_regression@sklearn.linear_model.linearregression.json'
          ).open() as file_pointer:
        got = json.load(file_pointer)

    assert got['pipeline_designator'] == 'tabular_regression@sklearn.linear_model.linearregression'


@pytest.fixture(scope='session')
def search_problem_def() -> ProblemDefinition:
    config = '''
    {
        "dataset": {
            "config": "memory",
            "column_roles": {
                "target": {
                    "name": "c"
                }
            },
            "params": {
                "train_data": "train_data",
                "test_data": "test_data"
            }
        },
        "problem_type": {
            "task": "regression"
        },
        "cross_validation": {
            "k": 2
        },
        "metrics" : {
            "root_mean_squared_error" : {}
        },
        "hyperparams": [
            {
                "select": {
                    "algorithm": "sklearn.ensemble.RandomForestRegressor"
                },
                "params": {
                    "n_estimators": {
                        "list": [100, 200]
                    }
                }
            }
        ]
    }
    '''

    return ProblemDefinition(
        clause=config)


def test_hyperparam_search(search_problem_def: ProblemDefinition) -> None:
    (train_data, test_data) = get_data()
    dut = Wrangler(
        problem_definition=search_problem_def,
        algorithm_catalog=FakeAlgorithmCatalogAuto)
    got = dut.fit_predict_rank()
    assert len(got.executable_pipelines) == 25
    assert {
        'tabular_regression@sklearn.linear_model.linearregression',
        'tabular_regression@sklearn.ensemble.randomforestregressor:'
        'max_depth=8,min_samples_split=10,n_estimators=100',
        'tabular_regression@sklearn.ensemble.randomforestregressor:'
        'max_depth=15,min_samples_split=10,n_estimators=200',
    }.issubset(got.executable_pipelines.keys())


@pytest.fixture(scope='session')
def order_problem_def() -> ProblemDefinition:
    config = '''
    {
        "dataset": {
            "config": "memory",
            "column_roles": {
                "target": {
                    "name": "c"
                }
            },
            "params": {
                "train_data": "train_data",
                "test_data": "test_data"
            }
        },
        "problem_type": {
            "task": "test_task"
        },
        "cross_validation": {
            "k": 3
        },
        "metrics" : {
            "root_mean_squared_error" : {}
        },
        "hyperparams": ["disable_grid_search"]
    }
    '''

    return ProblemDefinition(clause=config)


class FakeInst(FakeInstance):
    '''The instance of a fake algorithm for order tests.'''
    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        if dataset is None:
            return None
        retval = dataset.output()
        df = pd.DataFrame({'c': dataset.dataframe['a'] + 1})
        retval.predictions = df
        return retval


class FakeAlg(Algorithm):
    '''A fake algorithm used for order tests.'''
    _name = 'fake_algorithm'

    def instantiate(self, **hyperparams) -> 'FakeInstance':
        return FakeInst(parent=self, **hyperparams)


def test_order(order_problem_def) -> None:
    train_data = pd.DataFrame(
        {
            'a': range(1, 22),
            'b': range(1, 22),
            'c': range(2, 23)
        }
    )
    test_data = train_data  # noqa: F841 pylint: disable=unused-variable
    dut = Wrangler(
        problem_definition=order_problem_def)

    fake_template = PipelineTemplate(
        name='test_template',
        tags={
            'task': ['test_task'],
            'data_type': ['tabular']},
        generator=dut.generator
    )
    fake_template.step(model=FakeAlg())
    dut.template_catalog.register(fake_template)

    got = dut.fit_predict_rank()
    results = got.train_results.predictions[Designator('test_template')]
    assert results is not None
    assert results.predictions is not None
    predict_df = results.predictions
    assert got.split_dataset.ground_truth is not None
    ground_truth_df = got.split_dataset.ground_truth.ground_truth
    pd.testing.assert_frame_equal(predict_df[['c']], ground_truth_df)
    assert all(predict_df['c'] == predict_df['ground truth'])


def non_seeded_problem_def() -> ProblemDefinition:
    config = '''
    {
        "dataset": {
            "config": "memory",
            "column_roles": {
                "target": {
                    "name": "c"
                }
            },
            "params": {
                "train_data": "train_data",
                "test_data": "test_data"
            }
        },
        "problem_type": {
            "task": "test_task"
        },
        "cross_validation": {
            "k": 2
        },
        "metrics" : {
            "root_mean_squared_error" : {}
        },
        "hyperparams": ["disable_grid_search"]
    }
    '''
    train_data = pd.DataFrame(  # noqa: F841 pylint: disable=unused-variable
        {
            'a': range(1, 21),
            'b': range(1, 21),
            'c': range(2, 22)
        }
    )

    return ProblemDefinition(clause=config)


def seeded_problem_def(seed: int) -> ProblemDefinition:
    config = {
        "dataset": {
            "config": "memory",
            "column_roles": {
                "target": {
                    "name": "c"
                }
            },
            "params": {
                "train_data": "train_data",
                "test_data": "test_data"
            }
        },
        "problem_type": {
            "task": "test_task"
        },
        "cross_validation": {
            "k": 2
        },
        "metrics": {
            "root_mean_squared_error": {}
        },
        "hyperparams": [
            "disable_grid_search",
            {
                "select": {
                    "tags": {
                        "supports_random_seed": "true"
                    }
                },
                "params": {
                    "random_seed": {
                        "fixed": seed
                    }
                }
            }
        ]
    }
    train_data = pd.DataFrame(  # noqa: F841 pylint: disable=unused-variable
        {
            'a': range(1, 21),
            'b': range(1, 21),
            'c': range(2, 22)
        }
    )

    return ProblemDefinition(clause=config)


class RandomAlgorithmInstance(AlgorithmInstance):
    _rng: RandomState

    def __init__(self, parent: Algorithm, **hyperparams):
        super().__init__(parent)
        if 'random_seed' in hyperparams:
            self._rng = RandomState(hyperparams['random_seed'])

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        if dataset is None:
            return None
        retval = dataset.output()
        retval['predictions'] = pd.DataFrame({
            'c': list(self._rng.randint(0, 10, len(dataset.dataframe)))
        }).reset_index(drop=True)
        return retval


class RandomAlgorithm(Algorithm):
    _name = 'random_algorithm'
    _tags = {
        'task': ['test_task'],
        'data_type': ['tabular'],
        'supports_random_seed': ['true']
    }
    _default_hyperparams = {
        'random_seed': Defaults.SEED,
    }

    def instantiate(self, **hyperparams) -> RandomAlgorithmInstance:
        return RandomAlgorithmInstance(parent=self, **self.hyperparams(**hyperparams))


def test_set_seed() -> None:
    (train_data, test_data) = get_data()
    dut_def = Wrangler(problem_definition=non_seeded_problem_def())
    dut5678 = Wrangler(problem_definition=seeded_problem_def(5678))

    pipe_def = PipelineTemplate(
        name='test_template',
        tags={
            'task': ['test_task'],
            'data_type': ['tabular']},
        generator=dut_def.generator
    )
    pipe_def.step(model=RandomAlgorithm())
    dut_def.template_catalog.register(pipe_def)

    pipe5678 = PipelineTemplate(
        name='test_template',
        tags={
            'task': ['test_task'],
            'data_type': ['tabular']},
        generator=dut5678.generator
    )
    pipe5678.step(model=RandomAlgorithm())
    dut5678.template_catalog.register(pipe5678)

    got_def = dut_def.fit_predict_rank()
    got5678 = dut5678.fit_predict_rank()

    des_def = Designator('test_template')
    des5678 = Designator('test_template@random_algorithm:random_seed=5678')

    got_def_train_results_des_def = got_def.train_results[des_def]
    got5678_train_results_des5678 = got5678.train_results[des5678]

    assert got_def_train_results_des_def.prediction is not None
    assert got5678_train_results_des5678.prediction is not None

    predictions_def = got_def_train_results_des_def.prediction.predictions[['c']]
    predictions5678 = got5678_train_results_des5678.prediction.predictions[['c']]

    # The k-fold-cross-validator instantiates the algorithm separately
    # for each fold, so in this special case, we expect to see the same
    # sequence twice.

    want_def = pd.DataFrame({'c': [3, 3, 6, 2, 1, 8, 5, 1, 7, 3, 3, 3, 6, 2, 1, 8, 5, 1, 7, 3]})
    want5678 = pd.DataFrame({'c': [5, 3, 6, 9, 3, 3, 9, 8, 3, 3, 5, 3, 6, 9, 3, 3, 9, 8, 3, 3]})
    pd.testing.assert_frame_equal(predictions_def, want_def)
    pd.testing.assert_frame_equal(predictions5678, want5678)


@pytest.fixture(scope='session')
def disable_grid_search_problem_def() -> ProblemDefinition:
    config = '''
    {
        "dataset": {
            "config": "memory",
            "column_roles": {
                "target": {
                    "name": "c"
                }
            },
            "params": {
                "train_data": "train_data",
                "test_data": "test_data"
            }
        },
        "problem_type": {
            "task": "regression"
        },
        "cross_validation": {
            "k": 2
        },
        "metrics" : {
            "root_mean_squared_error" : {}
        },
        "hyperparams": [
            "disable_grid_search",
            {
                "select": {
                    "algorithm": "sklearn.ensemble.RandomForestRegressor"
                },
                "params": {
                    "n_estimators": {
                        "list": [100, 200],
                        "default": 100
                    }
                }
            }
        ],
        "output": {}
    }
    '''

    return ProblemDefinition(clause=config)


def test_disable_grid_search(disable_grid_search_problem_def: ProblemDefinition) -> None:
    (train_data, test_data) = get_data()
    dut = Wrangler(
        problem_definition=disable_grid_search_problem_def,
        algorithm_catalog=FakeAlgorithmCatalogAuto)
    got = dut.fit_predict_rank()
    assert {
        'tabular_regression@sklearn.linear_model.linearregression',
        'tabular_regression@sklearn.ensemble.randomforestregressor'
        ':max_depth=none,min_samples_split=2,n_estimators=100'
    } == set(got.executable_pipelines.keys())



'''Stub cross-validator does nothing'''
# pylint: disable=duplicate-code,unused-argument,too-many-arguments
from typing import Dict

from ..catalog.catalog import upcast
from ..executor.executor import Executor
from ..generator.bound_pipeline import BoundPipeline
from ..generator.designator import Designator
from ..instantiator.executable_pipeline import PipelineResult
from ..instantiator.instantiator_factory import InstantiatorFactory
from ..splitters.impl.splitter import SplitDataset
from ..wrangler.dataset import Dataset
from .impl.cross_validator import CrossValidator, CrossValidatorCatalog

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


class StubCrossValidator(CrossValidator):
    '''Stub cross-validator does nothing'''

    _name = 'stub_cross_validator'
    _tags = {
        'some_key': ['some_value'],
    }

    def validate_pipelines(self,
                           split_dataset: SplitDataset,
                           bound_pipelines: Dict[Designator, BoundPipeline],
                           instantiator: InstantiatorFactory,
                           executor: Executor,
                           **overrides
                           ) -> Dict[Designator, PipelineResult]:
        '''Stub cross-validator does nothing'''
        retval = {
            des: PipelineResult(
                prediction=Dataset(),
                bound_pipeline=pipe,
                split_dataset=split_dataset)
            for des, pipe in bound_pipelines.items()}
        return retval


def register(catalog: CrossValidatorCatalog, *args, **kwargs) -> None:
    '''Register all the objects in this file.'''
    val = StubCrossValidator()
    catalog.register(val, val.name, upcast(val.tags))



#!/bin/env python
"""Main for AutonML."""

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import argparse
import logging
import sys
import warnings
from typing import Callable, Optional

from ngautonml.algorithms.impl.algorithm_auto import AlgorithmCatalogAuto
from ngautonml.executor.simple.simple_executor import SimpleExecutor
from ngautonml.generator.generator import GeneratorImpl
from ngautonml.instantiator.instantiator_factory import InstantiatorFactory
from ngautonml.metrics.impl.metric_auto import MetricCatalogAuto
from ngautonml.problem_def.problem_def import ProblemDefinition
from ngautonml.ranker.ranker_impl import RankerImpl
from ngautonml.templates.impl.template_auto import TemplateCatalogAuto
from ngautonml.wrangler.logger import Level, Logger
from ngautonml.wrangler.wrangler import Wrangler
from ngautonml.wrangler.distributed_wrangler import DistributedWrangler

warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=UserWarning)


def set_global_log_level(args: argparse.Namespace) -> logging.Logger:
    '''Set log level for all of ngautonml, overridding levels set in each file.'''
    level: Optional[str] = args.log_level
    level_int: Optional[int] = None
    if level is not None:
        try:
            level_int = int(level)
        except ValueError:
            try:
                level_int = int(getattr(Level, level.upper()))
            except (ValueError, AttributeError) as exc:
                raise ValueError(
                    f'Log level {level} is not a valid level name and cannot be casted to int.  '
                    'Try one of: VERBOSE, DEBUG, INFO, WARN, ERROR.') from exc

    if level_int is not None:
        Logger.set_global_level(level=level_int)
    return Logger(__file__,
                  to_stdout=True,
                  level=level_int or Level.WARN).logger()


def wrangle(args: argparse.Namespace):
    """Do full autoML."""
    # TODO(piggy): Set up non-trivial architectural componenents
    # here based on command-line arguments.
    with open(args.problem_definition, 'r', encoding='utf8') as _f:
        problem_definition = ProblemDefinition(_f.read())
    log = set_global_log_level(args)
    log.info('%s', problem_definition)
    wrangler = Wrangler(
        problem_definition=problem_definition,
        metric_catalog=MetricCatalogAuto,
        algorithm_catalog=AlgorithmCatalogAuto,
        ranker=RankerImpl,
        template_catalog=TemplateCatalogAuto,
        generator=GeneratorImpl,
        executor=SimpleExecutor,
        instantiator_factory=InstantiatorFactory,
    )
    got = wrangler.fit_predict_rank()
    print(got.rankings)


def instantiate(args: argparse.Namespace):
    """Instantiate one or more pipelines.

    Optionally includes a trained model in the instantiated pipeline.
    """
    print(f'Instantiate: {args!r}')


def distributed(args: argparse.Namespace):
    """Stand up a distributed node."""
    with open(args.problem_definition, 'r', encoding='utf8') as _f:
        problem_definition = ProblemDefinition(_f.read())
    log = set_global_log_level(args)
    log.info('%s', problem_definition)
    wrangler = DistributedWrangler(
        problem_definition=problem_definition,
        my_id=args.my_id)
    wrangler.server(host=args.host, port=args.port)()


def usage(parser: argparse.ArgumentParser) -> Callable[
        [argparse.Namespace], None]:
    """Returns a function that prints the usage message."""
    def _f(_args: argparse.Namespace):
        parser.print_usage(sys.stderr)
    return _f


def main():
    """
    Just do something
    """
    # Create the top-level parser.
    parser_global = argparse.ArgumentParser(prog='picard')
    parser_global.set_defaults(func=usage(parser_global))
    parser_global.add_argument(
        '-l', '--log_level',
        type=str,
        help='Log level.  int or one of "VERBOSE", "DEBUG", "INFO", "WARN", "ERROR".',
        required=False
    )

    # Create sub-parser.
    sub_parsers = parser_global.add_subparsers(help='sub-command help')

    # Create the parser for the "wrangle" sub-command.
    parser_wrangle = sub_parsers.add_parser('wrangle',
                                            help='Run full autoML on a '
                                            'dataset.')
    parser_wrangle.set_defaults(func=wrangle)
    parser_wrangle.add_argument(
        '-d', '--problem_definition',
        type=str,
        help='path to problem definition JSON file',
        required=True)

    # Create the parser for the "instantiate" sub-command.
    parser_instantiate = sub_parsers.add_parser('instantiate',
                                                help='Generate an '
                                                'executable algorithm.')
    parser_instantiate.set_defaults(func=instantiate)
    parser_instantiate.add_argument(
        '-k', '--kind',
        action='append', choices=['d3m_json', 'python_script',
                                  'jupyter_notebook', 'airflow_class'],
        help='Indicate the kind of executable algorithm to generate. '
        'May be repeated. Defaults to all choices.')
    parser_instantiate.add_argument(
        '-p', '--pipeline',
        type=str,
        help='name or path to a pipeline to instantiate',
        required=True
    )
    parser_instantiate.add_argument(
        '-m', '--trained_model',
        action='append',
        help='optional path to the trained model to instantiate. '
        'May be repeated.'
    )

    parser_distributed = sub_parsers.add_parser('distributed',
                                                help='Start up a server running '
                                                'a set of distributed algorithms.')
    parser_distributed.set_defaults(func=distributed)
    parser_distributed.add_argument(
        '-d', '--problem_definition',
        type=str,
        help='path to problem definition JSON file',
        required=True)
    parser_distributed.add_argument(
        '-i', '--my_id',
        type=int,
        help='Node number of the server',
        required=False,
        default=1
    )
    parser_distributed.add_argument(
        '-H', '--host',
        type=str,
        help='Host name or IP address of web API',
        default='localhost',
    )
    parser_distributed.add_argument(
        '-P', '--port',
        type=int,
        help='Port number of web API',
        default=60080,
    )

    a = parser_global.parse_args()
    a.func(a)


if __name__ == "__main__":
    main()



'''Base class for abstractor instances.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
from typing import Dict, Optional, TypeVar

from ....algorithms.impl.fittable_algorithm_instance import UntrainedError
from ....neighbor_manager.node_id import NodeID
from ....wrangler.logger import Logger, Level
from ....wrangler.dataset import Dataset

from ...impl.distributed_algorithm_instance import DistributedAlgorithmInstance, NeighborState

log = Logger(__file__, to_stdout=False, level=Level.INFO).logger()


NeighborStateSubclass = TypeVar('NeighborStateSubclass', bound='AbstractorNeighborState')


class NoDataError(Exception):
    '''No data was provided to the abstractor.'''


class PredictCalledError(Exception):
    '''Predict was called on an abstractor.'''


class AbstractorNeighborState(NeighborState[NeighborStateSubclass]):
    '''A message from a neighbor in the abstractor network.'''

    @abc.abstractmethod
    def synthesize(self) -> Dataset:
        '''Synthesize a training dataset typical of the state.
        '''


class AbstractorInstance(DistributedAlgorithmInstance):
    '''Base class for abstractor instances.'''
    _neighbors: Dict[NodeID, AbstractorNeighborState]

    @property  # type: ignore[override]
    def _my_state(self) -> Optional[AbstractorNeighborState]:
        retval = DistributedAlgorithmInstance._my_state.fget(self)  # type: ignore[attr-defined] # pylint: disable=assignment-from-no-return,line-too-long
        assert retval is None or isinstance(retval, AbstractorNeighborState), (
            'BUG: expected _my_state to be None or an AbstractorNeighborState.')
        return retval

    @_my_state.setter
    def _my_state(self, value: Optional[AbstractorNeighborState]) -> None:
        assert value is None or isinstance(value, AbstractorNeighborState), (
            'BUG: expected value to be None or an AbstractorNeighborState.')
        DistributedAlgorithmInstance._my_state.fset(self, value)  # type: ignore[attr-defined]

    @property
    def my_state(self) -> Optional[AbstractorNeighborState]:
        '''Return the state of the Abstractor.'''
        return self._my_state

    def predict_passthrough(self, *args, **kwargs) -> Optional[Dataset]:
        '''Passthrough to the predict method of DistributedAlgorithm.'''
        return super().predict(*args, **kwargs)

    def predict(self, *args, **kwargs) -> Optional[Dataset]:
        '''Predict the output of the model.'''
        raise PredictCalledError('predict is not normally implemented in an Abstractor.')

    def _predict(self, *args, **kwargs) -> Optional[Dataset]:
        '''Predict the output of the model.'''
        raise PredictCalledError('predict is not normally implemented in an Abstractor.')

    @abc.abstractmethod
    def _fit(self, dataset: Optional[Dataset], **kwargs) -> None:
        '''Fit the model to the data. This is the actual implementation of fit.'''

    def synthesize(self) -> Dataset:
        '''Synthesize a training dataset typical of the state.'''
        my_state = self.my_state
        if my_state is None:
            raise UntrainedError('Cannot synthesize without a state.')
        return my_state.synthesize()



'''Handles for trained models.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pickle

from typing import Any, Dict, Optional

from ..algorithms.impl.algorithm import Algorithm
from ..algorithms.impl.algorithm_instance import AlgorithmInstance
from ..algorithms.impl.fittable_algorithm_instance import FittableAlgorithmInstance
from ..generator.designator import Designator, StepDesignator


class CucumberError(Exception):
    '''Base class for all errors thrown related to Cucumbers'''


class DeserializationError(CucumberError):
    '''Ran into issues while deserializing.'''


class Cucumber():
    '''Model-specific representation of a trained model.'''
    _serialized_model: Optional[bytes] = None
    _filename: StepDesignator
    _pipeline_designator: Designator
    _hyperparams: Dict[str, Any]
    _catalog_name: str

    @classmethod
    def deserialize(cls, pickled_cucumber: bytes) -> 'Cucumber':
        '''Convert a pickled Cucumber back into a Cucumber.'''
        return pickle.loads(pickled_cucumber)

    def __init__(self,
                 impl: AlgorithmInstance,
                 filename: StepDesignator,
                 pipeline_designator: Designator,
                 hyperparams: Dict[str, Any]):
        self._catalog_name = impl.catalog_name
        if isinstance(impl, FittableAlgorithmInstance):
            self._serialized_model = impl.serialize()
        self._filename = filename
        self._pipeline_designator = pipeline_designator
        self._hyperparams = hyperparams

    def serialize(self) -> bytes:
        '''Make an on-disk representation of the Cucumber.'''
        # TODO(Merritt/Piggy): Find more stable pickling protocol
        return pickle.dumps(self, protocol=pickle.HIGHEST_PROTOCOL)

    def deserialize_model(self, alg: Algorithm) -> AlgorithmInstance:
        '''Make a new instance of the algorithm passed in.

        Deserialize the saved trained model, and apply the saved hyperparams, if any.
        '''
        if alg.name != self._catalog_name:
            raise DeserializationError(
                f'Catalog name of Algorithm used for deserialization ({alg.name}) '
                f'and saved catalog name in cucumber ({self._catalog_name}) '
                'do not match.')
        hyperparams = self._hyperparams
        retval = alg.instantiate(**hyperparams)
        if isinstance(retval, FittableAlgorithmInstance):
            assert self._serialized_model is not None, (
                'BUG: cucumber asked to deserialize fittable model'
                ' but has no bytes to deserialize.')
            retval = retval.deserialize(serialized_model=self._serialized_model)
        assert isinstance(retval, AlgorithmInstance), (
            f'BUG: {alg.name}.instantiate() returned a  {type(retval)}, not an AlgorithmInstance '
            f'got: {retval}')
        return retval

    @property
    def serialized_model(self) -> Optional[bytes]:
        '''Typically a pickled representation of the trained model.'''
        return self._serialized_model

    @property
    def filename(self) -> StepDesignator:
        '''Returns the filename property.'''
        return self._filename

    @property
    def catalog_name(self) -> str:
        '''Returns the catalog name of the algorithm associated with this cucumber.'''
        return self._catalog_name

    @property
    def pipeline_designator(self) -> Designator:
        '''Returns the pipeline designator'''
        return self._pipeline_designator

    @property
    def hyperparams(self) -> Dict[str, Any]:
        '''Hyperparameter state for this cucumber.'''
        if self._hyperparams is None:
            return {}
        return self._hyperparams


class JarOfCucumbers(Dict[StepDesignator, Cucumber]):
    '''Trained models indexed by the designators for their steps.'''

    def trained_model(self, step_designator: StepDesignator) -> Cucumber:
        '''The trained model for the step designator.'''
        return self[step_designator]

    def set(self, designator: StepDesignator, trained_model: Cucumber) -> None:
        '''Saves the trained model into the training dictionary.'''
        self[designator] = trained_model



'''Tests for ParallelStep'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Any

from ...algorithms.impl.algorithm import Algorithm
from .parallel_step import ParallelStep
from .pipeline_template import PipelineTemplate

# pylint: disable=missing-function-docstring,missing-class-docstring,duplicate-code,protected-access


class ModelStub(Algorithm):
    def instantiate(self, **unused_kwargs) -> Any:
        return None


def test_parallel() -> None:
    dut = PipelineTemplate(name='Fake_Template')
    pipe1 = dut.new(name='at')
    step1a = pipe1.step(ModelStub('A'))
    step1t = pipe1.step(ModelStub('T'))
    pipe2 = dut.new(name='dt')
    step2d = pipe2.step(ModelStub('D'))
    step2t = pipe2.step(ModelStub('T'))
    parallel_step = dut.parallel(left=pipe1, right=pipe2)
    assert len(dut._steps) == 1
    assert isinstance(parallel_step, ParallelStep)
    assert set(parallel_step.subpipeline_keys) == set(['left', 'right'])
    assert parallel_step.subpipeline('left')._steps == [step1a, step1t]
    assert parallel_step.subpipeline('right')._steps == [step2d, step2t]



'''Tests for searcher.py'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Dict, List, Optional

from ..algorithms.connect import ConnectorModel
from ..algorithms.impl.algorithm import Algorithm
from ..algorithms.impl.algorithm_instance import AlgorithmInstance
from ..generator.bound_pipeline import BoundPipeline
from ..generator.designator import Designator
from ..problem_def.hyperparam_config import HyperparamConfig
from ..problem_def.task import TaskType
from ..templates.impl.pipeline_step import PipelineStep
from ..wrangler.constants import RangeMethod
from ..wrangler.dataset import Dataset
from .param_range import ParamRange
from .searcher import SearcherImpl, StepSet
from .frozen_overrides import FrozenOverrides

# pylint: disable=missing-docstring, duplicate-code


def test_step_set_fixed_sunny_day():
    # We use ConnectorModel just because its hyperparameters are easy to work with.
    step = PipelineStep(name='some_step',
                        model=ConnectorModel(overridable_param='overridable_value'))
    dut = StepSet(step=step)

    # 'multiply' binds all steps in the set.
    # it may increase the number of steps for RangeMethods other than FIXED,
    #   but that is not the case here.
    dut.multiply('some_param', ParamRange(RangeMethod.FIXED, prange='some_value'),
                 no_show=set())
    assert len(dut.steps) == 1
    new_step = list(dut.steps.values())[0]
    assert new_step.filename.endswith('@some_step:some_param=some_value')
    assert "connect_" in new_step.filename
    want_key = FrozenOverrides.freeze({
        new_step.pipeline_designator_component: {'some_param': 'some_value'}})
    assert dut.steps[want_key].hyperparams() == {
        'overridable_param': 'overridable_value',
        'some_param': 'some_value',
    }

    # Try another multiply and see that we get a new step.
    dut.multiply('another_param', ParamRange(RangeMethod.FIXED, 'another_value'),
                 no_show=set())
    assert len(dut.steps) == 1
    new_step = list(dut.steps.values())[0]
    assert new_step.filename.endswith(
        '@some_step:another_param=another_value,some_param=some_value')
    assert "connect_" in new_step.filename
    want_key = FrozenOverrides.freeze({new_step.pipeline_designator_component: {
        'some_param': 'some_value',
        'another_param': 'another_value'}})
    print(dut.steps)
    assert dut.steps[want_key].hyperparams() == {
        'overridable_param': 'overridable_value',
        'some_param': 'some_value',
        'another_param': 'another_value',
    }


def test_step_set_filenames():
    PipelineStep.reset_serial_number()
    step = PipelineStep(name='some_step',
                        model=ConnectorModel(overridable_param='overridable_value'))
    dut = StepSet(step=step)

    dut.multiply('some_param', ParamRange(RangeMethod.FIXED, prange='some_value'),
                 no_show=set())
    key1 = FrozenOverrides.freeze({'some_step': {'some_param': 'some_value'}})
    assert dut.steps[key1].pipeline_designator_component == 'some_step'
    assert dut.steps[key1].filename == '@connect_1@some_step:some_param=some_value'

    dut.multiply('overridable_param', ParamRange(RangeMethod.FIXED, prange='new_value'),
                 no_show=set())

    key2 = FrozenOverrides.freeze({'some_step': {
        'some_param': 'some_value',
        'overridable_param': 'new_value'}})
    assert dut.steps[key2].pipeline_designator_component == 'some_step'
    assert (
        dut.steps[key2].filename
        == '@connect_2@some_step:overridable_param=new_value,some_param=some_value')


def test_step_set_filenames_unnamed_step():
    PipelineStep.reset_serial_number()
    step = PipelineStep(model=ConnectorModel(overridable_param='overridable_value'))
    dut = StepSet(step=step)

    dut.multiply('some_param', ParamRange(RangeMethod.FIXED, prange='some_value'),
                 no_show=set())

    key1 = FrozenOverrides.freeze({
        'connect': {'some_param': 'some_value'}
    })
    assert dut.steps[key1].filename == '@connect_1@:some_param=some_value'


def test_step_set_list_sunny_day():
    step = PipelineStep(model=ConnectorModel(overridable_param='overridable_value'),
                        name='some_step')
    dut = StepSet(step=step)

    # This is the test.
    dut.multiply('some_param', ParamRange(RangeMethod.LIST, prange=['list_value1', 'list_value2']),
                 no_show=set())

    key1 = FrozenOverrides.freeze({'some_step': {'some_param': 'list_value1'}})
    key2 = FrozenOverrides.freeze({'some_step': {'some_param': 'list_value2'}})
    assert set(dut.steps.keys()) == {key1, key2}
    assert dut.steps[key1].hyperparams() == {
        'overridable_param': 'overridable_value',
        'some_param': 'list_value1',
    }
    # Can we append a FIXED value to each of those?
    dut.multiply('another_param', ParamRange(RangeMethod.FIXED, 'another_value'),
                 no_show=set())
    key3 = FrozenOverrides.freeze({'some_step': {
        'some_param': 'list_value2',
        'another_param': 'another_value'}})
    assert dut.steps[key3].hyperparams() == {
        'overridable_param': 'overridable_value',
        'some_param': 'list_value2',
        'another_param': 'another_value',
    }


def test_step_set_linear_sunny_day():
    step = PipelineStep(model=ConnectorModel(overridable_param='overridable_value'),
                        name='some_step')
    dut = StepSet(step=step)

    # This is the test.
    dut.multiply('some_param', ParamRange(RangeMethod.LINEAR, prange=[1, 1, 3]),
                 no_show=set())

    key1 = FrozenOverrides.freeze({'some_step': {'some_param': '1'}})
    key2 = FrozenOverrides.freeze({'some_step': {'some_param': '2'}})
    key3 = FrozenOverrides.freeze({'some_step': {'some_param': '3'}})
    assert set(dut.steps.keys()) == {key1, key2, key3}
    assert dut.steps[key3].hyperparams() == {
        'overridable_param': 'overridable_value',
        'some_param': 3,
    }
    # Can we append a FIXED value to each of those?
    dut.multiply('another_param', ParamRange(RangeMethod.FIXED, 'another_value'),
                 no_show=set())
    key4 = FrozenOverrides.freeze({'some_step': {
        'some_param': '2',
        'another_param': 'another_value'}})
    assert dut.steps[key4].hyperparams() == {
        'overridable_param': 'overridable_value',
        'some_param': 2,
        'another_param': 'another_value',
    }


def test_step_set_linear_complex():
    '''The elements of range for LINEAR hyperparams must be a subclass of Number.

    However, there is a special case of complex numbers where the < operation does not work.
    As this could potentially cause problems, ensure it is properly handled.
    '''
    step = PipelineStep(model=ConnectorModel(overridable_param='overridable_value'),
                        name='some_step')
    dut = StepSet(step=step)

    # This is the test.
    dut.multiply('some_param',
                 ParamRange(method=RangeMethod.LINEAR,
                            prange=[complex(1, 1), complex(1, 1), complex(3, 3)]),
                 no_show=set())

    key1 = FrozenOverrides.freeze({'some_step': {'some_param': '(1+1j)'}})
    key2 = FrozenOverrides.freeze({'some_step': {'some_param': '(2+2j)'}})
    key3 = FrozenOverrides.freeze({'some_step': {'some_param': '(3+3j)'}})
    assert set(dut.steps.keys()) == {key1, key2, key3}
    assert dut.steps[key3].hyperparams() == {
        'overridable_param': 'overridable_value',
        'some_param': complex(3, 3),
    }


FIXED_HYPERPARM_CONFIG = HyperparamConfig(clause={
    'hyperparams': [
        {
            'select': {
                'name': 'step2',
            },
            'params': {
                'added_param': {
                    'fixed': 8,
                },
                'new_param': {
                    'fixed': 16,
                }
            },
        },
    ],
})


def test_searcher_impl_fixed() -> None:
    pipeline = BoundPipeline(name='some_template')
    pipeline.step(model=ConnectorModel()).set_name('step1')
    step2 = pipeline.step(model=ConnectorModel(base_param=1),
                          added_param=2).set_name('step2')

    dut = SearcherImpl(FIXED_HYPERPARM_CONFIG)

    # This is the actual test.
    got: Dict[Designator, BoundPipeline] = dut.bind_hyperparams(pipeline=pipeline)

    # Confirm that we only got 1 pipeline back, because the parameters that we bound are FIXED.
    assert len(got) == 1
    new_pipeline = got[Designator('some_template@step2:added_param=8,new_param=16')]
    new_steps: List[PipelineStep] = new_pipeline.steps
    # Make sure the original step did not get modified.
    assert step2.hyperparams() == {
        'base_param': 1,
        'added_param': 2,
    }
    # Confirm that only the matched step got modified.
    assert new_steps[0].hyperparams() == {}
    assert new_steps[1].hyperparams() == {
        'base_param': 1,
        'added_param': 8,
        'new_param': 16,
    }


EMPTY_HYPERPARM_CONFIG = HyperparamConfig(clause={
    'hyperparams': []
})


def test_null_search() -> None:
    '''Test that Searcher behaves correctly when there are no matching overrides.'''
    pipeline = BoundPipeline(name='some_template')
    pipeline.step(model=ConnectorModel()).set_name('step1').mark_queried()
    pipeline.step(model=ConnectorModel(base_param=1)).set_name('step2')

    dut = SearcherImpl(EMPTY_HYPERPARM_CONFIG)
    got: Dict[Designator, BoundPipeline] = dut.bind_hyperparams(pipeline=pipeline)

    want_des = Designator('some_template@step1')
    assert len(got) == 1
    assert want_des in got
    assert got[want_des].designator == want_des


class FakeAlgorithmInstance(AlgorithmInstance):
    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        if dataset is None:
            return None
        return Dataset(dataset.copy())


class FakeAlgorithm(Algorithm):
    _name = 'fake_algorithm'
    _default_hyperparams = {
        'base_param': ParamRange(RangeMethod.LIST, ['a', 'b']),
        'param_to_override': 'value_to_override',
    }
    _tags = {
        'task': [
            TaskType.REGRESSION.name,
            TaskType.BINARY_CLASSIFICATION.name,
        ]
    }

    def instantiate(self, **hyperparams) -> FakeAlgorithmInstance:
        return FakeAlgorithmInstance(parent=self, **hyperparams)


def test_default_grid_search() -> None:
    pipeline = BoundPipeline('bound_pipeline')
    step = pipeline.step(model=FakeAlgorithm(param_to_override='overridden_value'))
    step.set_name('some_step')
    dut = SearcherImpl(EMPTY_HYPERPARM_CONFIG)

    # This is the test.
    got = dut.bind_hyperparams(pipeline)

    assert set(got.keys()) == {
        'bound_pipeline@some_step:base_param=a',
        'bound_pipeline@some_step:base_param=b'
    }
    assert got[Designator('bound_pipeline@some_step:base_param=a')].steps[0].hyperparams() == {
        'base_param': 'a',
        'param_to_override': 'overridden_value'
    }
    assert got[Designator('bound_pipeline@some_step:base_param=b')].steps[0].hyperparams() == {
        'base_param': 'b',
        'param_to_override': 'overridden_value'
    }


def test_non_matching_overrides() -> None:
    '''When a searcher is created with only Override(s) whose Selector(s) do not match any steps
        of the pipeline, we want it to return a single pipeline that is a copy of what we
        started with.
    '''
    pipeline = BoundPipeline('bound_pipeline')
    step = pipeline.step(model=ConnectorModel(some_param='some_value'))
    step.set_name('some_step')
    dut = SearcherImpl(HyperparamConfig({
        'hyperparams': [
            {
                'select': {
                    'algorithm': 'does not match'
                },
                'params': {
                    'some_param': {
                        'fixed': 'a_value_we_wont_see'
                    }
                }
            },
            {
                'select': {
                    'algorithm': 'also_does_not_match',
                },
                'params': {
                    'some_param': {
                        'fixed': 'another_value_we_wont_see',
                    }
                },
            },
        ]
    }))

    # This is the test.
    got = dut.bind_hyperparams(pipeline)

    assert set(got.keys()) == {
        'bound_pipeline'
    }
    assert len(got[Designator('bound_pipeline')].steps) == 1



'''A model that returns output identical to input.

Used for parallel subpipelines where one of them has no operations in it.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Optional
from ..algorithms.impl.algorithm import Algorithm, MemoryAlgorithmCatalog
from ..algorithms.impl.algorithm_instance import AlgorithmInstance
from ..catalog.catalog import upcast
from ..wrangler.dataset import Dataset


class IdentityModelInstance(AlgorithmInstance):
    '''A model that returns output identical to input.'''

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        if dataset is None:
            return None
        new_dataset = dataset.output()
        new_dataset.update(dataset)
        return new_dataset


class IdentityModel(Algorithm):
    '''A model that returns output identical to input.'''
    _name = 'identity'
    _instance_constructor = IdentityModelInstance
    _tags = {
        'source': ['auton_lab'],
        'preprocessor': ['true'],
    }


def register(catalog: MemoryAlgorithmCatalog, *args, **kwargs) -> None:
    '''Register all the objects in this file.'''
    model = IdentityModel(*args, **kwargs)
    catalog.register(model, model.name, upcast(model.tags))



'''Test PipelineTemplate object.'''
from typing import Any

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


from ...algorithms.impl.algorithm import Algorithm
from .pipeline_template import PipelineTemplate

# pylint: disable=missing-function-docstring,missing-class-docstring,duplicate-code,protected-access


class ModelStub(Algorithm):
    def instantiate(self, **unused_kwargs) -> Any:
        return None


def test_instantiate() -> None:
    dut = PipelineTemplate(name='Fake_Template')
    assert dut.name == 'fake_template'


def test_step() -> None:
    dut = PipelineTemplate(name='Fake_Template')
    step_a = dut.step(ModelStub('A'))
    step_t = dut.step(ModelStub('T'))
    assert dut._steps == [step_a, step_t]



'''Decider that uses the algorithm-specific distance metric.'''
# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from enum import Enum
import math
import time
from typing import Mapping, Optional

from aenum import Enum as AEnum  # type: ignore[import-untyped]

from ..algorithms.impl.neighbor_metadata_interface import NeighborMetadataInterface
from ..config_components.distributed_config import DecidersConfig, DeciderConfig
from ..neighbor_manager.node_id import NodeID  # type: ignore[import-untyped]

from .impl.decider import Decider
from .impl.decider_catalog import DeciderCatalog


class DistanceDeciderConfig(DeciderConfig):
    '''Configuration for DistanceDecider.'''

    def __init__(self, name: str, clause: dict) -> None:
        super().__init__(name=name, clause=clause)

    def __str__(self) -> str:
        return f'threshold: {self.threshold}'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DistanceDeciderConfig):
            return False
        return self.threshold == other.threshold

    class Defaults(Enum):
        '''Default values for the clause.'''
        DEFAULT_THRESHOLD = 0.5
        GEOMETRIC_TERM = 1.0
        LINEAR_TERM = 0.0
        MIN = 0.0

    class Keys(AEnum):
        '''Keys for the clause.'''
        ENABLED = 'enabled'
        THRESHOLD = 'threshold'
        GEOMETRIC_TERM = 'geometric_term'
        LINEAR_TERM = 'linear_term'
        MIN = 'min'
        INTERVAL_SECONDS = 'interval_seconds'

    @property
    def threshold(self) -> float:
        '''Get the threshold.

        This is the initial value for the threshold.
        '''
        return self._get_with_default(
            self.Keys.THRESHOLD, dflt=self.Defaults.DEFAULT_THRESHOLD.value)

    @property
    def geometric_term(self) -> float:
        '''Get the geometric term.'''
        return self._get_with_default(
            self.Keys.GEOMETRIC_TERM, dflt=self.Defaults.GEOMETRIC_TERM.value)

    @property
    def linear_term(self) -> float:
        '''Get the linear term.'''
        return self._get_with_default(
            self.Keys.LINEAR_TERM, dflt=self.Defaults.LINEAR_TERM.value)

    @property
    def min(self) -> float:
        '''Get the minimum value.'''
        return self._get_with_default(
            self.Keys.MIN, dflt=self.Defaults.MIN.value)

    @property
    def interval_seconds(self) -> Optional[float]:
        '''Get the interval in seconds.

        If this is None (the default), every time the decider is called,
        the threshold will advance one cycle. This is cycle time.

        If this is not None, we advance by wall time.

        Thresholds are calculated one decide call before they are applied,
        so the initial threshold is always used at least once.
        '''
        return self._get_with_default(
            self.Keys.INTERVAL_SECONDS, dflt=None)


class DistanceDecider(Decider):
    '''Decider that uses the algorithm-specific distance metric.'''
    name = 'distance'
    tags = {}
    _config: DistanceDeciderConfig

    def _threshold(self, my_meta: NeighborMetadataInterface) -> float:
        '''Get the threshold, calculating the next threshold for later.'''
        if self.name not in my_meta.scratch:
            my_meta.scratch[self.name] = {
                'threshold': self._config.threshold,
                'start_time': time.monotonic(),
                'cycle': 1,
            }
        retval = my_meta.scratch[self.name]['threshold']

        assert self._config is not None

        initial_threshold = self._config.threshold
        geom_term = self._config.geometric_term
        linear_term = self._config.linear_term
        min_val = self._config.min
        interval_seconds = self._config.interval_seconds

        threshold = my_meta.scratch[self.name]['threshold']
        if interval_seconds is not None:
            # Advance by wall time.
            elapsed = time.monotonic() - my_meta.scratch[self.name]['start_time']
            intervals = math.ceil(elapsed / interval_seconds)
        else:
            # Advance by cycle.
            intervals = my_meta.scratch[self.name]['cycle']
            my_meta.scratch[self.name]['cycle'] += 1
        threshold = initial_threshold * (geom_term ** intervals) + (intervals * linear_term)
        threshold = max(threshold, min_val)
        my_meta.scratch[self.name]['threshold'] = threshold
        return retval

    def decide(self, my_meta: NeighborMetadataInterface,
               neighbor_id: NodeID, neighbors: Mapping[NodeID, NeighborMetadataInterface]) -> bool:
        '''Decide if we should send to neighbor.'''
        assert my_meta.current_state is not None
        threshold = self._threshold(my_meta)
        if neighbors[neighbor_id].current_state is None:
            # We haven't heard from this neighbor yet. Let them know our state.
            return True
        return (my_meta.current_state.distance(neighbors[neighbor_id].current_state)
                > threshold)  # type: ignore[union-attr]


def register(catalog: DeciderCatalog) -> None:
    '''Register all objects in this file.'''
    catalog.register(DistanceDecider)
    DecidersConfig.register(name=DistanceDecider.name, config_type=DistanceDeciderConfig)



'''A Table representing a pandas DataFrame.'''

# Copyright (c) 2025 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Any, Dict, List, Union

import numpy as np
import pandas as pd

from .impl.table import TableFactory, Table
from .impl.table_catalog import TableCatalog


class DataFrameTable(Table):
    '''A Table representing a pandas DataFrame.'''
    name = 'dataframe'
    _value: pd.DataFrame
    my_type = pd.DataFrame

    def to_csv(self, *args, **kwargs) -> str:
        '''Output the table as a CSV string.'''
        return self._value.to_csv(*args, **kwargs)

    def as_dict(self, *args, **kwargs) -> Dict[Union[str, int], Any]:
        '''Convert table to a dictionary.'''
        return self._value.to_dict(*args, **kwargs)

    def drop(self, columns: List[Union[int, str]]) -> Table:
        '''Return a copy of this table with selected columns dropped.'''
        return TableFactory(self._value.drop(columns, axis=1))

    @property
    def empty(self) -> bool:
        '''Return True if the table is empty.'''
        return self._value.empty

    @property
    def columns(self) -> List[Union[int, str]]:
        '''Return the column names of the table.'''
        return list(self._value.columns)

    @columns.setter
    def columns(self, value: List[Union[int, str]]) -> None:
        '''Set the column names of the table.'''
        self._value.columns = pd.Index(value)


def as_numpy(from_table: Table) -> np.ndarray:
    '''Export a DataFrameTable to a numpy array.'''
    assert isinstance(from_table, DataFrameTable)
    if from_table.shape[1] == 1:
        return from_table.value().to_numpy().flatten()
    return from_table.value().to_numpy()


def register(catalog: TableCatalog):
    '''Register the DataFrameTable with the catalog.'''
    catalog.register(DataFrameTable)
    TableFactory.register_constructor(pd.DataFrame, DataFrameTable)
    TableFactory.register_exporter(
        from_type=pd.DataFrame, to_type=pd.DataFrame, exporter=DataFrameTable.value)
    TableFactory.register_exporter(
        from_type=pd.DataFrame, to_type=np.ndarray, exporter=as_numpy)



'''Tests for plugin_catalog.py.'''

import os
from pathlib import Path
import subprocess

import pytest

from .catalog import Catalog
from .plugin_catalog import PluginCatalog
from .widgets.impl.widget_catalog import NamedWidget

# pylint: disable=missing-function-docstring,redefined-outer-name,duplicate-code

EXTRA_WIDGET_SOURCE = '''
from typing import Any, Optional
from ngautonml.catalog.widgets.impl.widget_catalog import NamedWidget, WidgetCatalog

class ExtraWidget(NamedWidget):
    _name = 'Gizmo'
    _tags = {
        'some_tag': ['home_directory']
    }


def register(catalog: WidgetCatalog):
    catalog.register(ExtraWidget())
'''


@pytest.fixture(scope="session")
def home_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    retval = tmp_path_factory.mktemp("data")
    source_path = retval / '.ngautonml' / 'plugins' / 'widgets' / 'extra_widget.py'
    source_path.parent.mkdir(parents=True, exist_ok=True)
    with source_path.open('wt') as file:
        file.write(EXTRA_WIDGET_SOURCE)
    return retval


def test_home_directory_auto(home_path) -> None:
    os.environ['HOME'] = str(home_path)
    dut: Catalog = PluginCatalog('widgets', default_root=Path(__file__).parent)
    got1 = dut.lookup_by_name('Gizmo')
    assert isinstance(got1, NamedWidget)
    # if this were using the Gizmo in ./widgets/some_subfolder/gizmo.py,
    #   the value of 'some_tag' would be 'test_dir'
    assert got1.tags['some_tag'] == ['home_directory']

    got2 = dut.lookup_by_name('DooDad')
    assert isinstance(got2, NamedWidget)
    # we are using the DooDad in ./widgets/doodad.py, so the
    #   value of 'some_tag_ is 'test_dir'
    assert got2.tags['some_tag'] == ['test_dir']
    assert hasattr(got2, 'quux')


def test_with_plugin() -> None:
    pluginpath = (Path(__file__).parents[2] / 'plugins' / 'testplugin'
                  / 'dist' / 'testplugin-0.0.1-py3-none-any.whl')
    subprocess.run(['python', '-m', 'build'], cwd=pluginpath.parents[1], check=True)
    subprocess.run(['pip', 'install', pluginpath], check=False)
    try:
        dut: Catalog = PluginCatalog('widgets', default_root=Path(__file__).parent)
        got = dut.lookup_by_name('DooDad')
        assert isinstance(got, NamedWidget)
        # because testplugin is installed, DooDad will be overridden by the one
        #   in testplugin, where the value of 'some_tag' is 'plugin_dir'
        assert got.tags['some_tag'] == ['plugin_dir']
        assert hasattr(got, 'quux')
    finally:
        subprocess.run(['pip', 'uninstall', '-y', 'testplugin'], check=False)


def test_register_to_plugin_catalog(home_path) -> None:
    os.environ['HOME'] = str(home_path)
    dut: PluginCatalog = PluginCatalog('widgets', default_root=Path(__file__).parent)

    new_doodad = NamedWidget(name='DooDad', tags={'some_tag': ['hand_registered']})

    dut.register(new_doodad)

    got = dut.lookup_by_name('DooDad')
    assert isinstance(got, NamedWidget)
    assert got.tags['some_tag'] == ['hand_registered']



'''Widgets for testing.'''

from ...memory_catalog import MemoryCatalog
from ...catalog_element_mixin import CatalogElementMixin


class Widget():
    '''This is the thing that is stored in WidgetCatalog.'''


class NamedWidget(Widget, CatalogElementMixin):
    '''Widget that properly inherits from CatalogElementMixin and thus has a name'''


class WidgetCatalog(MemoryCatalog[Widget]):
    '''This is the class we will exercise.'''



'''Base class for distributed linear algorithms.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
from typing import Callable, Iterable, List, Optional, Tuple, TypeVar, Union

import numpy as np
import pandas as pd
from scipy.optimize import minimize  # type: ignore[import]


from ...algorithms.impl.algorithm import Algorithm, HyperparamError
from ...config_components.distributed_config import DistributedConfig
from ...tables.impl.table import TableFactory
from ...wrangler.dataset import Dataset, RoleName
from ...wrangler.logger import Level, Logger

from .distributed_algorithm_instance import (
    DistributedAlgorithmInstance, NeighborState, NoDataError)


log = Logger(__file__, level=Level.INFO).logger()


# These functions assume x is 1 or 2 dimensional and v is 1-dimensional


def grad_norm2(a: np.ndarray) -> np.ndarray:
    '''Return the gradient of the squared norm of a.'''
    return 2.0 * a


def grad_inner_product(a: np.ndarray, b: np.ndarray) -> np.ndarray:  # pylint: disable=unused-argument
    '''Return the gradient of the inner product of a and b.'''
    return a


def inner_product(a: np.ndarray, b: np.ndarray) -> Union[np.float64, np.ndarray]:
    '''Return the inner product of a and b.'''
    return a @ b


def norm2(a: np.ndarray) -> np.float64:
    '''Return the squared norm of a.'''
    return np.square(a).sum()  # np.linalg.norm(a) ** 2


NeighborStateSubclass = TypeVar('NeighborStateSubclass', bound='NeighborState')


class LinearNeighborState(NeighborState[NeighborStateSubclass]):
    '''Base class for Linear Neighbor States.'''
    _v: np.ndarray

    def __init__(self, v: np.ndarray, columns: Optional[List[str]] = None) -> None:
        super().__init__(columns=columns)
        self._v = v

    def __str__(self) -> str:
        '''Used for more readable logs when investigating convergence.'''
        return str(self._v)

    def __repr__(self) -> str:
        '''A more verbose representation of our state than __str__.'''
        return f'{self.__class__.__name__}(v = {self._v}, columns = {self.columns})'

    @property
    def v(self) -> np.ndarray:
        '''The internal state of the model.'''
        return self._v

    @v.setter
    def v(self, value: np.ndarray) -> None:
        '''The internal state of the model.'''
        self._v = value

    def _distance(self, other: 'LinearNeighborState') -> np.float64:
        '''Compute distance between two models.'''
        assert isinstance(other, type(self))
        a: np.float64 = norm2(other.v)
        b: np.float64 = norm2(self.v)
        ab = inner_product(self.v, other.v)
        assert isinstance(ab, np.float64), (
            f'BUG: expected np.float64, instead found {type(ab)}'
        )
        dist: np.float64 = a + b - 2 * ab
        return dist

    def distance(self, other: NeighborStateSubclass) -> float:
        assert isinstance(other, type(self))
        return float(self._distance(other))


class LinearDistributedAlgorithmInstance(DistributedAlgorithmInstance, metaclass=abc.ABCMeta):
    '''Base class for distributed linear algorithms.'''

    _l2: np.float64
    _tol: np.float64
    _maxiter: int

    def __init__(self,
                 parent: Algorithm,
                 distributed: DistributedConfig,
                 **kwargs):
        hyperparams = parent.hyperparams(**kwargs)
        if not isinstance(distributed, DistributedConfig):
            raise HyperparamError(
                'distributed must be a DistributedConfig, instead found '
                f'{distributed} of type {type(distributed)}')
        self._l2 = np.float64(hyperparams.pop('L2'))

        self._last_dataset = None
        super().__init__(parent, distributed=distributed, **hyperparams)

    @property
    def _neighbor_models_iter(self) -> Iterable[LinearNeighborState]:
        for v in super()._neighbor_models_iter:
            assert isinstance(v, LinearNeighborState), (
                'BUG: expected neighbor_models to contain AutonLinearRegressionNeighbor, '
                f'instead found {v} of type {type(v)}'
            )
            yield v

    @property
    @abc.abstractmethod
    def _yhat_is_proba(self) -> bool:
        '''Is yhat probabilites?'''

    @property
    def _yhat_is_multivalued(self) -> bool:
        '''True if yhat returns a vector of results, 1 per column of x.'''
        return False

    @abc.abstractmethod
    def _prepare_data(self, dataset: Optional[Dataset]) -> Tuple[np.ndarray, np.ndarray]:
        '''Prepare data for fitting or predicting.'''
        raise NotImplementedError

    def _standardize_y(self, y: np.ndarray) -> np.ndarray:
        '''Regularize the target values.'''
        log.error('%s._standardize_y should not be called.', type(self).__name__)
        return y

    def _objective_with_gradient(
        self, x: np.ndarray, y: np.ndarray, v_old: np.ndarray,
        l2: np.float64, _lambda: np.float64, omega: np.float64,
        k: int) -> Tuple[Callable[[np.ndarray], np.float64],
                         Callable[[np.ndarray], np.ndarray]]:
        '''Return the objective function and its gradient.'''
        _, _, _, _, _, _, _ = x, y, v_old, l2, _lambda, omega, k
        log.error(
            '%s._objective_with_gradient should not be called.', type(self).__name__)
        return (lambda _: np.float64(0.0), lambda _: np.zeros_like(v_old))

    @abc.abstractmethod
    def _yhat(self, x: np.ndarray) -> np.ndarray:
        '''Return the predicted values as an np.ndarray.'''
        raise NotImplementedError

    def _fit(self, dataset: Optional[Dataset], **kwargs) -> None:
        '''Fit a model based on train data.

        This sets self.trained to True.
        '''

        try:
            np_cov, np_tar = self._prepare_data(dataset)
        except NoDataError:
            return  # There is nothing to do.

        log.log(5,  # log level verbose
                'Node %s (%x) is fitting %s; len(neighbor_models): %s,\n dataset: %r',
                self._my_id, id(self),
                type(self).__name__, len(list(self._neighbor_models_iter)), dataset)
        x = np_cov
        n, d = x.shape

        # If we have neither data nor neighbors, there's nothing to do.
        if dataset is None and not list(self._neighbor_models_iter):
            return

        omega = self._omega
        _lambda = self._lambda

        assert self._my_state is not None, (
            'BUG: self._my_state should not be None when _fit is called.'
        )
        assert isinstance(self._my_state, LinearNeighborState), (
            'BUG: expected self._my_state to be a LinearNeighborState, '
            f'instead found {self._my_state} of type {type(self._my_state)}'
        )
        y = self._standardize_y(np_tar)
        k = len(self._neighbor_metadata)

        if n == 0:
            # If you have no data take the mean of your neighbors.
            # If there are also no neighbors do nothing.
            if k != 0:
                self._my_state.v = np.mean([m.v for m in self._neighbor_models_iter], axis=0)
        else:
            if self._my_state.v.size != d:
                # no previous fit
                # v_old (v from last fit) to all zeros
                v_old = np.zeros(d)
                # set omega to 1 to disable self-regularization
                omega = np.float64(1.0)
            else:
                # previous fit exists
                v_old = self._my_state.v.copy()

            l2 = self._l2
            fun, jac = self._objective_with_gradient(
                x=x, y=y, v_old=v_old, l2=l2, _lambda=_lambda, omega=omega, k=k)
            x0 = self._my_state.v if self._my_state is not None else np.zeros(d)
            if self._maxiter is not None:
                soln = minimize(
                    fun, x0, method='BFGS', jac=jac,
                    tol=self._tol, options={'maxiter': self._maxiter})
            else:
                soln = minimize(
                    fun, x0, method='BFGS', jac=jac,
                    tol=self._tol)
            self._my_state.v = soln.x

    def _predict(self, dataset: Optional[Dataset], **kwargs) -> Optional[Dataset]:
        '''Apply model to input dataset to create output.

        This may require that the model is fit (self.trained == True) before it is called.
        '''
        assert self._predict_state is not None, (
            'BUG: self._predict_state should not be None when predict is called.'
        )
        assert isinstance(self._predict_state, LinearNeighborState), (
            'BUG: expected my_state to be a LinearNeighborState, '
            f'instead found {self._predict_state} of type {type(self._predict_state)}'
        )

        if dataset is None:
            return None

        x = self._prepare_data(dataset)[0]

        yhat = self._yhat(x)

        retval = dataset.output()
        # TODO(Piggy/Merritt): Convert to picard format.
        #   Preserve the column names as class names
        #   Other classification algorithms should follow the
        #   format of one column per class with probabilities
        #   for that class.
        if self._yhat_is_multivalued:
            pred_df = pd.DataFrame(yhat).transpose()
            assert self._columns is not None
            pred_df.columns = pd.Index(self._columns)
        else:
            target_col_name = dataset.metadata.roles[RoleName.TARGET][0].name
            if self._yhat_is_proba:
                retval.probabilities = TableFactory(pd.DataFrame(yhat))  # type: ignore[abstract] # pylint: disable=abstract-class-instantiated

                pred_df = pd.DataFrame({
                    target_col_name: np.argmax(yhat, axis=1)
                })
            else:
                pred_df = pd.DataFrame({
                    target_col_name: yhat
                })
        retval.predictions = pred_df
        return retval

    @property
    def coef(self) -> np.ndarray:
        '''Weights for the model.'''
        assert self._predict_state is not None
        assert isinstance(self._predict_state, LinearNeighborState)
        return self._predict_state.v[:-1].copy()

    @property
    def intercept(self) -> np.float64:
        '''Intercept for the model.'''
        assert self._predict_state is not None
        assert isinstance(self._predict_state, LinearNeighborState)
        return self._predict_state.v[-1]

    @property
    def params(self) -> np.ndarray:
        '''All the params for the model.'''
        assert self._predict_state is not None
        assert isinstance(self._predict_state, LinearNeighborState)
        return self._predict_state.v



'''
Catalog for pipeline templates

pipeline templates are DAGs of "slots" that can contain either a specific algorithm,
or a query to the AlgorithmCatalog for the Generator to fill in.
'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Dict, List, Union

from ...catalog.catalog import CatalogError
from ...catalog.memory_catalog import MemoryCatalog
from ...problem_def.task import Task
from .pipeline_template import PipelineTemplate


class TemplateRegisterError(CatalogError):
    '''Template missing required keys for registration.'''


class TemplateCatalog(MemoryCatalog[PipelineTemplate]):
    '''Mixin class for pipeline template catalogs'''
    def validate(self):
        '''Validate the catalog is capable of generating a pipeline.'''
        errors = []

        for obj in self.all_objects():
            final_tags = obj.tags
            # Check that at least one task and data type exists
            tasks = final_tags[Task.Keys.TASK_TYPE.value]  # pylint: disable=no-member,line-too-long
            data_types = final_tags[Task.Keys.DATA_TYPE.value]  # pylint: disable=no-member,line-too-long

            if 0 == len(tasks):
                errors.append(TemplateRegisterError(
                    f'No task defined in Template Register for {obj.name}'))

            if 0 == len(data_types):
                errors.append(TemplateRegisterError(
                    f'No data type defined in Template Register for {obj.name}'))

        if 0 < len(errors):
            raise TemplateRegisterError(errors)

    # TODO(piggy): Rewrite this in terms of Task.
    def lookup_by_task(self, task: Union[str, List[str]]) -> Dict[str, PipelineTemplate]:
        '''Returns list of all templates that match at least one given task'''
        results: Dict[str, PipelineTemplate] = {}

        tasks = task

        if isinstance(task, str):
            tasks = [task]

        for tas in tasks:
            results.update(
                self.lookup_by_tag_and(**{Task.Keys.TASK_TYPE.value: tas}))  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
        return results

    # TODO(piggy): Rewrite this in terms of DataType.
    def lookup_by_datatype(self, data_type: Union[str, List[str]]) -> Dict[str, PipelineTemplate]:
        '''Returns list of all templates that match at least one given task'''
        results: Dict[str, PipelineTemplate] = {}

        data_types = data_type

        if isinstance(data_type, str):
            data_types = [data_type]

        for dts in data_types:
            results.update(
                self.lookup_by_tag_and(**{Task.Keys.DATA_TYPE.value: dts}))  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long

        return results

    # TODO(piggy): Rewrite this in terms of ProblemType.
    def lookup_by_both(self,
                       data_type: Union[str, List[str]],
                       task: Union[str, List[str]]
                       ) -> Dict[str, PipelineTemplate]:
        '''Returns all templates that match any combination of a data_type and task'''
        by_datatype = self.lookup_by_datatype(data_type)
        by_task = self.lookup_by_task(task)

        return dict(by_datatype.items() & by_task.items())


class TemplateCatalogStub(TemplateCatalog):
    '''stub'''



'''A designator identifying a specific step or pipeline.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


class Designator(str):
    '''A unique designator identifying a specific pipeline.

    It is also used in the filename when saving that pipeline to disk.
    '''
    def __str__(self) -> str:
        return self.lower()


class StepDesignator(str):
    '''A unique designator identifying a specific step in a specific pipeline.

    It is also used in the filename when saving the model in that step to disk.'''
    def __str__(self) -> str:
        return self.lower()



'''Decider that uses the algorithm-specific distance metric to compare against our last state.'''
# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Mapping

from ..algorithms.impl.neighbor_metadata_interface import NeighborMetadataInterface
from ..config_components.distributed_config import DecidersConfig
from ..neighbor_manager.node_id import NodeID

from .distance import DistanceDeciderConfig, DistanceDecider
from .impl.decider_catalog import DeciderCatalog


class SelfDistanceDeciderConfig(DistanceDeciderConfig):
    '''Configuration for SelfDistanceDecider.'''


class SelfDistanceDecider(DistanceDecider):
    '''Decider that uses the algorithm-specific distance metric against our last state.'''
    name = 'self_distance'
    tags = {}

    def decide(self, my_meta: NeighborMetadataInterface,
               neighbor_id: NodeID, neighbors: Mapping[NodeID, NeighborMetadataInterface]) -> bool:
        '''Decide if we should send to neighbor.'''
        if my_meta.current_state is None:
            return False
        if my_meta.last_state is None:
            return True
        threshold = self._threshold(my_meta)
        return (my_meta.current_state.distance(my_meta.last_state_sent)
                > threshold)  # type: ignore[union-attr]


def register(catalog: DeciderCatalog) -> None:
    '''Register all objects in this file.'''
    catalog.register(SelfDistanceDecider)
    DecidersConfig.register(name=SelfDistanceDecider.name, config_type=SelfDistanceDeciderConfig)



'''Catalog for objects that load bound pipelines.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
from typing import Dict, List, Optional, Union

from ...algorithms.impl.algorithm import AlgorithmCatalog
from ...catalog.memory_catalog import MemoryCatalog
from ...templates.impl.template import TemplateCatalog

from .pipeline_loader import PipelineLoader


class PipelineLoaderCatalog(MemoryCatalog[PipelineLoader], metaclass=abc.ABCMeta):
    '''Base class for PipelineLoader catalogs'''
    _algorithm_catalog: Optional[AlgorithmCatalog]
    _template_catalog: Optional[TemplateCatalog]

    def __init__(self,
                 algorithm_catalog: Optional[AlgorithmCatalog] = None,
                 template_catalog: Optional[TemplateCatalog] = None,
                 **kwargs):
        super().__init__(**kwargs)
        self._algorithm_catalog = algorithm_catalog
        self._template_catalog = template_catalog

    def register(self, obj: PipelineLoader, name: Optional[str] = None,
                 tags: Optional[Dict[str, Union[str, List[str]]]] = None) -> str:
        return super().register(obj, name, tags)


class PipelineLoaderCatalogStub(PipelineLoaderCatalog):
    '''stub for PipelineLoaderCatalog'''



'''Decider that requires all of its subdeciders to decide True.'''
# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Mapping

from ..algorithms.impl.neighbor_metadata_interface import NeighborMetadataInterface
from ..config_components.distributed_config import DecidersConfig
from ..neighbor_manager.node_id import NodeID

from .impl.decider_catalog import DeciderCatalog

from .boolean import BooleanDeciderConfig, BooleanDecider


class AndDeciderConfig(BooleanDeciderConfig):
    '''Configuration for AndDecider.'''

    def __str__(self) -> str:
        return f'and({",".join([str(c) for c in self._clause])})'


class AndDecider(BooleanDecider):
    '''Decider that decides True if all of its subdeciders decide True.'''
    name = 'and'
    tags = {}
    _config: AndDeciderConfig

    def decide(self, my_meta: NeighborMetadataInterface,
               neighbor_id: NodeID, neighbors: Mapping[NodeID, NeighborMetadataInterface]) -> bool:
        '''Decide if we should send to neighbor.'''
        return all(self._all_decisions(
            my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors))


def register(catalog: DeciderCatalog) -> None:
    '''Register all objects in this file.'''
    catalog.register(AndDecider)
    DecidersConfig.register(name=AndDecider.name, config_type=AndDeciderConfig)



'''Aggregator object class for configurable and user-made aggregators.'''
import abc

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Any, Dict, List, Optional, Union

from ...catalog.catalog_element_mixin import CatalogElementMixin
from ...generator.designator import Designator
from ...instantiator.executable_pipeline import PipelineResults
from ...metrics.impl.metric import Metric, MetricStub
from ...ranker.ranker import Ranking, Rankings, RankingStub, ScoredResult
from ...wrangler.dataset import Dataset


class AggregatorError(Exception):
    '''Base error class for aggregators'''


class AggregatorNameError(AggregatorError):
    '''Aggregator lacks a proper indexing name.'''


class AggregatorAttributeError(AggregatorError):
    '''Aggregator is missing a required attribute.'''


class AggregatorInvalidDatasetError(AggregatorError):
    '''Aggregator was passed a dataset it cannot use.'''


class AggregatorInsufficientPipelines(AggregatorError):
    '''Aggregator requires more pipelines.'''


class Aggregator(Metric, CatalogElementMixin, metaclass=abc.ABCMeta):
    '''Base class wrapper for accessing Aggregator objects.'''
    _hyperparams: Optional[Dict[str, Any]] = None

    def __init__(self,
                 name: Optional[str] = None,
                 tags: Optional[Dict[str, List[str]]] = None,
                 **overrides):
        super().__init__(name=name, tags=tags)
        if self._hyperparams is None:
            self._hyperparams = {}
        params = self._hyperparams.copy()
        params.update(overrides)
        self._hyperparams = params

    def __str__(self):
        return self.name

    @abc.abstractmethod
    def aggregate(self, rankings: Rankings, all_scores: bool = False) -> Ranking:
        '''Calculates the Aggregator results.'''

    def calculate(self, pred: Dataset, ground_truth: Optional[Dataset] = None) -> float:
        '''Calculates the metric results.'''
        raise AggregatorError('calculate is not implemented for Aggregators')

    def hyperparams(self, **overrides) -> Dict[str, Any]:
        '''Get the hyperparameters for this aggregator.

        Arguments passed to this function override defaults for the aggregator.
        '''
        assert self._hyperparams is not None, (
            'BUG: Aggregator._hyperparams should be resolved in __init__.')
        hyperparams = self._hyperparams.copy()
        hyperparams.update(**overrides)
        return hyperparams


class AggregatorStub(Aggregator):
    '''This is a stub'''
    _name = 'stub_aggregator'
    _high = True

    def aggregate(self, rankings: Rankings, all_scores: bool = False) -> Ranking:
        return RankingStub(metric=MetricStub(self._name), results=PipelineResults())


class AggregatorRanking(Ranking):
    '''This is the Aggregator version of a Ranking.

    We pass the scores in at init time.
    '''

    def __init__(self, metric: Aggregator, results: PipelineResults,
                 scores: Union[Dict[Designator, int], str]):
        super().__init__(metric=metric, results=results)
        self._scores = []
        if isinstance(scores, str):
            self._scores = [ScoredResult(metric=metric, result=result, unscorable_reason=scores)
                            for result in results.values()]
        else:
            for des, ranking in scores.items():
                scored_result = ScoredResult(metric=metric, result=results[des], score=ranking)
                self._scores.append(scored_result)
            self._scores.sort(reverse=metric.high)

    def score(self) -> None:
        '''Scores are precalculated (do nothing).'''



'''Base class for classes that load pipelines.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
from typing import Any, Dict, Optional

from ngautonml.algorithms.impl.algorithm import AlgorithmCatalog
from ngautonml.templates.impl.template import TemplateCatalog

from ...generator.bound_pipeline import BoundPipeline


class Error(BaseException):
    '''base class for all errors thrown by this module'''


class MissingArgumentsError(Error):
    '''The pipeline loader did not encounter a required argument.'''


class UnknownArgumentsError(Error):
    '''The pipeline loader encounter unknown arguments to load()'''


class PipelineLoader(metaclass=abc.ABCMeta):
    '''Base class for classes that load pipelines.

    '''
    name: str
    tags: Dict[str, Any]

    _algorithm_catalog: Optional[AlgorithmCatalog]
    _template_catalog: Optional[TemplateCatalog]

    def __init__(self,
                 algorithm_catalog: Optional[AlgorithmCatalog] = None,
                 template_catalog: Optional[TemplateCatalog] = None,
                 **unused_kwargs):
        self._algorithm_catalog = algorithm_catalog
        self._template_catalog = template_catalog
        super().__init__()

    def load(self, *args, **kwargs) -> BoundPipeline:
        '''Load the pipeline.'''
        # This exists so we have a place to hang logging or other
        # monitoring.
        return self._load(*args, **kwargs)

    @abc.abstractmethod
    def _load(self, *args, **kwargs) -> BoundPipeline:
        '''Load the pipeline.'''


class PipelineLoaderStub(PipelineLoader):
    '''stub'''
    name = 'stub_pipeline_loader'
    tags: Dict[str, Any] = {}

    def _load(self, *args, **unused_kwargs) -> BoundPipeline:
        return BoundPipeline(name=args[0])



'''wrapper of sklearn.model_selection.KFold'''
# pylint: disable=duplicate-code,too-many-locals

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from sklearn.model_selection import KFold  # type: ignore[import]

from ..problem_def.cross_validation_config import CrossValidationConfig
from ..splitters.impl.splitter import Splitter, SplitterCatalog, SplitDataset, Fold
from ..wrangler.dataset import Dataset


class KFoldSplitter(Splitter):
    '''Splitter that splits dataset into k folds'''

    _name = 'sklearn.model_selection.KFold'
    _hyperparams = {
        'n_splits': None,
    }
    _tags = {
        'default': ['true']
    }

    def split(self, dataset: Dataset, **overrides) -> SplitDataset:
        hyperparams = self.hyperparams(**overrides)
        n_splits = hyperparams['n_splits'] or get_num_splits(dataset.dataframe.shape[0])

        data_df = dataset.dataframe
        kfcv = KFold(
            n_splits=n_splits,
            shuffle=True,
            random_state=self._cv_config.seed)

        folds = []
        for train_index, test_index in kfcv.split(data_df):
            validate = dataset.output()
            train = dataset.output()
            ground_truth = dataset.output()

            train_df = data_df.iloc[train_index]
            validate_df = data_df.iloc[test_index]

            if dataset.metadata.target is not None:
                validate_df_target = validate_df[[dataset.metadata.target.name]]
                validate_df_no_target = validate_df.drop(dataset.metadata.target.name, axis=1)
            else:
                validate_df_target = None
                validate_df_no_target = validate_df

            train_df.reset_index(inplace=True, drop=True)
            validate_df_no_target.reset_index(inplace=True, drop=True)
            validate_df_target.reset_index(inplace=True, drop=True)

            train.dataframe = train_df
            validate.dataframe = validate_df_no_target
            ground_truth.ground_truth = validate_df_target

            folds.append(Fold(train=train, validate=validate, ground_truth=ground_truth))

        return SplitDataset(folds)


def register(catalog: SplitterCatalog, *args, cv_config: CrossValidationConfig, **kwargs):
    '''Register all the objects in this file.'''
    catalog.register(KFoldSplitter(*args, cv_config=cv_config, **kwargs))


def get_num_splits(length):
    ''' Determine number of splits based on the size of the dataset. '''
    splits = 2

    if length < 10000:
        splits = 5
    elif length < 20000:
        splits = 3

    return splits



''' A model that takes a dataset and returns only that columns that have

a specific role (most commonly, attribute and target)'''
from typing import Any, List, Optional, Union
import pandas as pd

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ..algorithms.impl.algorithm import Algorithm, MemoryAlgorithmCatalog
from ..algorithms.impl.algorithm_instance import AlgorithmInstance
from ..catalog.catalog import upcast
from ..wrangler.dataset import Dataset, RoleName


class Error(BaseException):
    '''Base class for all errors thrown by this module'''


class InvalidColumnSplitError(Error):
    '''Unable to split columns as specified.'''


class ExtractColumnsByRoleModelInstance(AlgorithmInstance):
    '''Returns only columns with a specific role.'''
    _desired_roles: List[RoleName]

    def __init__(self, parent,
                 desired_roles: Optional[Union[RoleName, List[RoleName]]] = None):
        super().__init__(parent=parent)
        assert desired_roles is not None

        if isinstance(desired_roles, RoleName):
            desired_roles = [desired_roles]

        self._desired_roles = desired_roles

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        if dataset is None:
            return None

        retval = dataset.output()
        retval_data = {}
        for desired_role in self._desired_roles:
            if desired_role not in dataset.roles.keys():
                continue

            columns_satisfying_role = dataset.roles[desired_role]
            for col in columns_satisfying_role:
                if col.name not in dataset.get_dataframe():
                    continue
                retval_data[col.name] = dataset.get_dataframe()[col.name]

        retval.dataframe = pd.DataFrame(retval_data)

        return retval


class ExtractColumnsByRoleModel(Algorithm):
    '''Returns only columns with a specific role.'''
    _name: str = 'Extract Columns by Role'
    _instance_constructor = ExtractColumnsByRoleModelInstance
    _tags = {
        'source': ['auton_lab'],
        'preprocessor': ['true'],
    }

    def _param_from_json(self, hyperparam_name: str, json_value: Any) -> Any:
        if hyperparam_name == 'desired_roles':
            if isinstance(json_value, str):
                return RoleName[json_value]
            assert isinstance(json_value, list)
            return [RoleName[role] for role in json_value]
        return super()._param_from_json(hyperparam_name, json_value)

    def param_to_json(self, hyperparam_name: str, python_value: Any) -> Any:
        if hyperparam_name == 'desired_roles':
            if isinstance(python_value, RoleName):
                return python_value.value
            assert isinstance(python_value, list)
            assert isinstance(python_value[0], RoleName)
            return [role.value for role in python_value]
        return super().param_to_json(hyperparam_name, python_value)


def register(catalog: MemoryAlgorithmCatalog, *args, **kwargs) -> None:
    '''Register all the objects in this file.'''
    model = ExtractColumnsByRoleModel(*args, **kwargs)
    catalog.register(model, model.name, upcast(model.tags))



'''An instantiator for the "simple" executor.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.
from typing import Dict
from pathlib import Path

from ...instantiator.instantiator import Instantiator
from ...generator.bound_pipeline import BoundPipeline
from ...generator.designator import StepDesignator
from .simple_executable_pipeline import SimpleExecutablePipeline


class SimpleInstantiator(Instantiator):
    '''An instantiator for the "simple" executor.'''

    def instantiate(self, pipeline: BoundPipeline) -> SimpleExecutablePipeline:
        retval = SimpleExecutablePipeline(
            pipeline=pipeline)
        return retval

    def save(self, pipeline: BoundPipeline, model_paths: Dict[StepDesignator, Path]) -> Path:
        raise NotImplementedError('Serialization does not make sense for SimpleExecutable.')
        # Fight me.



'''Step object for SimpleExecutor'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Iterable, Optional

from ...algorithms.impl.algorithm import Algorithm
from ...algorithms.impl.algorithm_instance import AlgorithmInstance
from ...algorithms.impl.fittable_algorithm_instance import FittableAlgorithmInstance
from ...generator.designator import Designator
from ...templates.impl.pipeline_step import PipelineStep
from ...wrangler.dataset import Dataset
from ..cucumber import Cucumber, JarOfCucumbers


class Error(Exception):
    '''Base error for SimpleStep'''


class InstantiationError(Error):
    '''Something has gone wrong while attempting to instantiate a model.'''


class SimpleExecutableStep(PipelineStep):
    '''Step representation for the SimpleExecutor

    The constructor 'compiles' a BoundStep (really a PipelineStep)
    into a SimpleStep.
    '''
    _bound_step: PipelineStep
    _model_instance: Optional[AlgorithmInstance]
    _trained = False

    def __init__(self,
                 bound_step: PipelineStep,
                 model: Optional[Algorithm] = None):
        super().__init__(
            model=model or bound_step._model,
            **bound_step.hyperparams())
        model = self._model
        self._bound_step = bound_step
        self._overrides = bound_step._overrides
        self._name = bound_step.opt_name
        self._serial_number = bound_step.serial_number
        self._init_model()

    def _init_model(self):
        '''Provide a customization point for model handling.

        There are some steps such as ParallelStep which need to
        override this because they don't have models.
        '''
        if self._model is None:
            raise InstantiationError(
                f'Can not instantiate {self._bound_step.designator} as model is not set.')
        instance = self._model.instantiate(**self._bound_step.hyperparams())
        if self._bound_step.serialized_model is not None:
            instance = instance.deserialize(self._bound_step.serialized_model)
            self.set_trained()
        self._model_instance = instance

    def set_trained(self):
        '''Force a step to the trained state.'''
        self._trained = True

    @property
    def locked(self) -> bool:
        '''Does this step have a model instance that is currently locked?'''
        return False if self._model_instance is None else self._model_instance.locked

    def converged(self) -> bool:
        '''If this step has a distributed model, has it converged?

        Always returns True for steps with no models or a non-distributed model.

        The threshold is the number of fits since the model changed enough to send to neighbors.
        '''
        return False if self._model_instance is None else self._model_instance.converged

    def fit(self, *args, **kwargs) -> None:
        '''Fit a model based on train data.

        The default fit method does nothing.
        '''
        assert self._model_instance is not None
        if isinstance(self._model_instance, FittableAlgorithmInstance):
            self._model_instance.fit(*args, **kwargs)

    def predict(self, *args, **kwargs) -> Optional[Dataset]:
        '''Apply model to input dataset to create output.

        This may require that the model is fit (self.trained == True) before it is called.
        '''
        assert self._model_instance is not None
        return self._model_instance.predict(*args, **kwargs)

    def cucumberize_all(self, pipeline_designator: Designator) -> JarOfCucumbers:
        '''Make this step into a Cucumber and put it in a JarOfCucumbers.

        Parallel steps will override this, but in the non-parallel case,
            this step only contains one algorithm and thus yields a jar of
            1 cucumber.
        '''
        assert self._model_instance is not None, (
            'BUG: cucumberize() called when model instance is none')
        assert self.filename is not None, (
            'BUG: cucumberize() called when filename is none')
        return JarOfCucumbers({
            self.filename:
            Cucumber(
                impl=self._model_instance,
                filename=self.filename,
                pipeline_designator=pipeline_designator,
                hyperparams=self.hyperparams()
            )}
        )

    @property
    def trained(self) -> bool:
        '''Has this model been trained?'''
        if isinstance(self._model_instance, FittableAlgorithmInstance):
            return self._model_instance.trained
        return True

    def start(self) -> None:
        '''Start the model.'''
        assert self._model_instance is not None
        self._model_instance.start()

    def stop(self) -> None:
        '''Stop the model.'''
        assert self._model_instance is not None
        self._model_instance.stop()

    @property
    def all_instances(self) -> Iterable[AlgorithmInstance]:
        '''Iterate through all algorithm instances.'''
        assert self._model_instance is not None
        yield self._model_instance



'''Tests for time_since_sent_specific.py'''
# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import time
from typing import Optional


from ..algorithms.impl.neighbor_metadata import NeighborMetadata
from ..algorithms.impl.distributed_algorithm_instance import NeighborState
from ..config_components.distributed_config import DistributedConfig
from ..deciders.impl.decider import Decider
from ..neighbor_manager.node_id import NodeID

from .impl.decider_auto import DeciderCatalogAuto

# pylint: disable=duplicate-code


class TestNeighborState(NeighborState):
    '''Test class for NeighborState.'''
    _center: float

    def __init__(self, center: float = 0.0) -> None:
        super().__init__()
        self._center = center

    @property
    def center(self) -> float:
        '''Get the center.'''
        return self._center

    def encode(self) -> bytes:
        '''Encode a message for distributed neighbors. '''
        return b'test neighbor state'

    @classmethod
    def decode(cls, serialized_model: bytes) -> 'TestNeighborState':
        '''Decode a message from distributed neighbors.

        This should just redirect to the relevant NeighborState subclass.
        '''
        return cls(center=0.0)

    def state_differs(self,
                      distributed: DistributedConfig,
                      other: Optional[NeighborState]) -> bool:
        '''Decide if our state differs meaningfully from the last one we sent.

        :distributed: contains information from the problem definition, including fit_eps
        :last_state_sent: is an instance of type(self) or None.
        '''
        return True

    def distance(self, other: NeighborState) -> float:
        '''Return the distance between two states.'''
        assert isinstance(other, TestNeighborState)
        return abs(self.center - other.center)


def test_time_since_sent_to() -> None:
    '''Test the time_since_sent_to decider.'''
    distributed = DistributedConfig(
        clause={
            'decider': {
                'time_since_sent_to': {
                    'time_seconds': 1.0,
                }
            },
        }
    )
    catalog = DeciderCatalogAuto()
    decider: Decider = catalog.lookup_by_name('time_since_sent_to')(
        config=distributed.decider.time_since_sent_to)  # type: ignore[attr-defined] # pylint: disable=no-member

    my_state = TestNeighborState()
    my_meta = NeighborMetadata(my_state=my_state)

    neighbor_id2 = NodeID(2)
    neighbor_id3 = NodeID(3)
    neighbors = {
        neighbor_id2: NeighborMetadata(my_state=TestNeighborState()),
        neighbor_id3: NeighborMetadata(my_state=TestNeighborState())
    }

    # We send the first time.
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id2, neighbors=neighbors) is True
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id3, neighbors=neighbors) is True
    # Mark that we sent to both neighbors.
    neighbors[neighbor_id2].last_state_sent = my_meta.current_state
    neighbors[neighbor_id3].last_state_sent = my_meta.current_state
    # We should not send immediately after sending...
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id2, neighbors=neighbors) is False
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id3, neighbors=neighbors) is False
    time.sleep(0.1)
    # ...or after a short interval.
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id2, neighbors=neighbors) is False
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id3, neighbors=neighbors) is False
    time.sleep(0.5)
    # Refresh node 3.
    neighbors[neighbor_id3].last_state_sent = my_meta.current_state
    time.sleep(0.5)
    # We should send to node 2 after a full interval, but not to node 3.
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id2, neighbors=neighbors) is True
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id3, neighbors=neighbors) is False
    neighbors[neighbor_id2].last_state_sent = my_meta.current_state
    # The nodes are now out of sync.
    time.sleep(0.1)
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id2, neighbors=neighbors) is False
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id3, neighbors=neighbors) is False
    time.sleep(0.5)
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id2, neighbors=neighbors) is False
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id3, neighbors=neighbors) is True
    neighbors[neighbor_id3].last_state_sent = my_meta.current_state
    time.sleep(0.5)
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id2, neighbors=neighbors) is True
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id3, neighbors=neighbors) is False



'''Test the DiscovererCatalog'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from .discoverer import DiscovererStub
from .discoverer_catalog import DiscovererCatalogStub

# pylint: disable=missing-function-docstring,duplicate-code


def test_register() -> None:
    dut = DiscovererCatalogStub()
    assert 'stub_discoverer' == dut.register(obj=DiscovererStub)


def test_register_name() -> None:
    dut = DiscovererCatalogStub()
    stub_discoverer = DiscovererStub
    assert 'somenameddiscoverer' == dut.register(
        obj=stub_discoverer, name='SomeNamedDiscoverer')


def test_lookup_by_name() -> None:
    dut = DiscovererCatalogStub()
    some_discoverer = DiscovererStub
    dut.register(obj=some_discoverer, name='SomeNamedDiscoverer')
    assert dut.lookup_by_name('SomeNamedDiscoverer') == some_discoverer


def test_lookup_one_by_tag() -> None:
    dut = DiscovererCatalogStub()
    dut.register(obj=DiscovererStub, name='NamedStubDiscoverer', tags={'k1': ['v1']})
    assert len(dut.lookup_by_tag_and(k1='v1')) == 1


def test_lookup_by_tag_fail() -> None:
    dut = DiscovererCatalogStub()
    dut.register(obj=DiscovererStub, name='NamedStubDiscoverer', tags={'k1': ['v1']})
    assert len(dut.lookup_by_tag_and(k1='different_value')) == 0


def test_lookup_multi_by_tag() -> None:
    dut = DiscovererCatalogStub()
    stub_discoverer = DiscovererStub
    dut.register(obj=stub_discoverer, name="Name1", tags={'k1': ['v1']})
    stub_discoverer2 = DiscovererStub
    dut.register(obj=stub_discoverer2, name="Name2", tags={'k1': ['v1']})
    stub_discoverer3 = DiscovererStub
    dut.register(obj=stub_discoverer3, name="Name3", tags={'k1': ['different_value']})
    assert len(dut.lookup_by_tag_and(k1='v1')) == 2



'''Decider that requires its subdecider to decide False.'''
# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from copy import deepcopy
from typing import Mapping

from ..algorithms.impl.neighbor_metadata_interface import NeighborMetadataInterface
from ..config_components.distributed_config import DecidersConfig, DeciderConfig
from ..neighbor_manager.node_id import NodeID

from .impl.decider_catalog import DeciderCatalog

from .boolean import BooleanDeciderConfig, BooleanDecider


class NotDeciderConfig(BooleanDeciderConfig):
    '''Configuration for NotDecider.'''

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        clause = deepcopy(self._clause)
        clause.pop(DeciderConfig.Keys.ENABLED, None)
        if len(clause) != 1:
            raise ValueError('NotDeciderConfig takes only one subclause.')

    def __str__(self) -> str:
        return f'not({",".join([str(c) for c in self._clause])})'


class NotDecider(BooleanDecider):
    '''Decider that decides True if its subdecider decides False.'''
    name = 'not'
    tags = {}
    _config: NotDeciderConfig

    def decide(self, my_meta: NeighborMetadataInterface,
               neighbor_id: NodeID, neighbors: Mapping[NodeID, NeighborMetadataInterface]) -> bool:
        '''Decide if we should send to neighbor.'''
        return not all(self._all_decisions(
            my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors))


def register(catalog: DeciderCatalog) -> None:
    '''Register all objects in this file.'''
    catalog.register(NotDecider)
    DecidersConfig.register(name=NotDecider.name, config_type=NotDeciderConfig)



'''A step representing a query to the model catalog

The generator will build a bound pipeline for every model that fits the query.
This type of step cannot exist in bound pipelines.
'''
import logging
from typing import Dict, Iterable, List, Union

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ...algorithms.impl.algorithm import AlgorithmCatalog, Algorithm

from ...generator.designator import Designator
from .pipeline_step import PipelineStep


class QueryStep(PipelineStep):
    '''A step representing a query to the model catalog.'''

    _algorithm_catalog: AlgorithmCatalog
    _query_tags: Dict[str, Union[str, Iterable[str]]]

    def __init__(self,
                 algorithm_catalog: AlgorithmCatalog,
                 **query_tags: Union[str, Iterable[str]]):
        super().__init__()
        self._query_tags = query_tags
        self._algorithm_catalog = algorithm_catalog

    def generate(self,
                 future_steps: List['PipelineStep']
                 ) -> Dict[Designator, List['PipelineStep']]:
        '''Generate pipeline sketches (lists of bound steps) for this step and all subsequent steps.

        This and subsequent query steps may expand into multiple 'pipeline sketches'.

        Handles the QueryStep case.
        '''
        models: Dict[str, Algorithm] = self._algorithm_catalog.lookup_by_tag_and(**self._query_tags)
        # Only use distributed algrithms if they are explicitly requested.
        if 'distributed' not in self._query_tags or 'true' not in self._query_tags['distributed']:
            models = {
                k: m for k, m in models.items()
                if 'distributed' not in m.tags or 'false' in m.tags['distributed']
            }
        logging.info('query found %d models with tags: %r', len(models), self._query_tags)
        new_steps = {
            Designator(model_name): [
                PipelineStep(model=model).set_name(model_name).mark_queried()
            ]
            for model_name, model in models.items()
        }
        if len(future_steps) == 0:
            return new_steps

        step0 = future_steps[0]
        sketches = step0.generate(future_steps[1:])
        retval: Dict[Designator, List['PipelineStep']] = {}
        for _, model_step in new_steps.items():
            for future_key, future_rest in sketches.items():
                retval[Designator(
                    f'{model_step[0].pipeline_designator_component}:{future_key}')] = (
                    model_step + future_rest[:])
        return retval



'''Wraps TSAD aggregators with TSADAggregator and registers them.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


from typing import Any, Dict, List
from .impl.aggregator_catalog import AggregatorCatalog
from .tsad_aggregator import TSADAggregator


AGGREGATORS: List[Dict[str, Any]] = [
    {
        'name': 'tsad.rank_aggregation.borda',
        'tags': {'for_tests': ['true']},
    },
    {
        'name': 'tsad.rank_aggregation.kemeny',
        'tags': {'required_number_of_pipelines': ['3']},
    },
    {
        'name': 'tsad.rank_aggregation.partial_borda',
        'weights': None,
        'top_k': 5,
    },
    {
        'name': 'tsad.rank_aggregation.trimmed_borda',
        'weights': None,
        'top_k': 5,
        'top_kr': None,
        'aggregation_type': 'kemeny',
        'metric': 'influence',
        'n_neighbors': 6,
    },
    {
        'name': 'tsad.rank_aggregation.trimmed_partial_borda',
        'tags': {'for_tests': ['true']},
        'weights': None,
        'top_k': 5,
        'top_kr': None,
        'aggregation_type': 'kemeny',
        'metric': 'influence',
        'n_neighbors': 6,
    },
]


def register(catalog: AggregatorCatalog, *unused_args, **unused_kwargs) -> None:
    '''Register all the objects in this file.'''
    for agg in AGGREGATORS:
        catalog.register(TSADAggregator(**agg))
